"""
Mocks for docstrings
"""

from itertools import chain
from textwrap import indent

from cdd.shared.pure_utils import identity, tab

docstring_header_no_nl_str = (
    "Acquire from the official tensorflow_datasets model zoo,"
    " or the ophthalmology focussed ml-prepare"
)
docstring_header_str = "{docstring_header_no_nl_str}\n".format(
    docstring_header_no_nl_str=docstring_header_no_nl_str
)

_docstring_header_and_return_str = (
    ":return: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))",
    ":rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```",
)

docstring_header_and_return_str = "\n".join(
    (docstring_header_str, *_docstring_header_and_return_str)
)

docstring_header_and_return_no_nl_str = "\n".join(
    (docstring_header_no_nl_str, *_docstring_header_and_return_str)
)

docstring_header_and_return_two_nl_str = "\n".join(
    (docstring_header_no_nl_str, "\n", *_docstring_header_and_return_str)
)

docstring_extra_colons_str = """
Some comment

:param dataset_name: Example: foo
:type dataset_name: ```str```
"""

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/losses.py#L845-L858
docstring_google_tf_squared_hinge_no_args_doc_str = (
    "Initializes `SquaredHinge` instance."
)
docstring_google_tf_squared_hinge = (
    docstring_google_tf_squared_hinge_no_args_doc_str,
    "Args:",
    "  reduction: (Optional) Type of `tf.keras.losses.Reduction` to apply to",
    "    loss. Default value is `AUTO`. `AUTO` indicates that the reduction",
    "    option will be determined by the usage context. For almost all cases",
    "    this defaults to `SUM_OVER_BATCH_SIZE`. When used with",
    "    `tf.distribute.Strategy`, outside of built-in training loops such as",
    "    `tf.keras` `compile` and `fit`, using `AUTO` or `SUM_OVER_BATCH_SIZE`",
    "    will raise an error. Please see this custom training [tutorial](",
    "      https://www.tensorflow.org/tutorials/distribute/custom_training)",
    "    for more details.",
    "  name: Optional name for the op. Defaults to 'squared_hinge'.",
)
docstring_google_tf_squared_hinge_str = "\n".join(docstring_google_tf_squared_hinge)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/optimizer_v2/adam.py#L35-L103
docstring_google_tf_adam = (
    "Optimizer that implements the Adam algorithm.",
    "Adam optimization is a stochastic gradient descent method that is based on",
    "adaptive estimation of first-order and second-order moments.",
    "According to",
    "[Kingma et al., 2014](http://arxiv.org/abs/1412.6980),",
    'the method is "*computationally',
    "efficient, has little memory requirement, invariant to diagonal rescaling of",
    "gradients, and is well suited for problems that are large in terms of",
    'data/parameters*".',
    "Args:",
    "learning_rate: A `Tensor`, floating point value, or a schedule that is a",
    "  `tf.keras.optimizers.schedules.LearningRateSchedule`, or a callable",
    "  that takes no arguments and returns the actual value to use, The",
    "  learning rate. Defaults to 0.001.",
    "beta_1: A float value or a constant float tensor, or a callable",
    "  that takes no arguments and returns the actual value to use. The",
    "  exponential decay rate for the 1st moment estimates. Defaults to 0.9.",
    "beta_2: A float value or a constant float tensor, or a callable",
    "  that takes no arguments and returns the actual value to use, The",
    "  exponential decay rate for the 2nd moment estimates. Defaults to 0.999.",
    "epsilon: A small constant for numerical stability. This epsilon is",
    '  "epsilon hat" in the Kingma and Ba paper (in the formula just before',
    "  Section 2.1), not the epsilon in Algorithm 1 of the paper. Defaults to",
    "  1e-7.",
    "amsgrad: Boolean. Whether to apply AMSGrad variant of this algorithm from",
    '  the paper "On the Convergence of Adam and beyond". Defaults to `False`.',
    "name: Optional name for the operations created when applying gradients.",
    '  Defaults to `"Adam"`.',
    "**kwargs: Keyword arguments. Allowed to be one of",
    '  `"clipnorm"` or `"clipvalue"`.',
    '  `"clipnorm"` (float) clips gradients by norm; `"clipvalue"` (float) clips',
    "  gradients by value.",
    "  ",
    "Usage:",
    ">>> opt = tf.keras.optimizers.Adam(learning_rate=0.1)",
    ">>> var1 = tf.Variable(10.0)",
    ">>> loss = lambda: (var1 ** 2)/2.0       # d(loss)/d(var1) == var1",
    ">>> step_count = opt.minimize(loss, [var1]).numpy()",
    ">>> # The first step is `-learning_rate*sign(grad)`",
    ">>> var1.numpy()",
    "9.9",
    "Reference:",
    "- [Kingma et al., 2014](http://arxiv.org/abs/1412.6980)",
    "- [Reddi et al., 2018](",
    "    https://openreview.net/pdf?id=ryQu7f-RZ) for `amsgrad`.",
    "Notes:",
    "The default value of 1e-7 for epsilon might not be a good default in",
    "general. For example, when training an Inception network on ImageNet a",
    "current good choice is 1.0 or 0.1. Note that since Adam uses the",
    "formulation just before Section 2.1 of the Kingma and Ba paper rather than",
    'the formulation in Algorithm 1, the "epsilon" referred to here is "epsilon',
    'hat" in the paper.',
    "The sparse implementation of this algorithm (used when the gradient is an",
    "IndexedSlices object, typically because of `tf.gather` or an embedding",
    "lookup in the forward pass) does apply momentum to variable slices even if",
    "they were not used in the forward pass (meaning they have a gradient equal",
    "to zero). Momentum decay (beta1) is also applied to the entire momentum",
    "accumulator. This means that the sparse behavior is equivalent to the dense",
    "behavior (in contrast to some momentum implementations which ignore momentum",
    "unless a variable slice was actually used).",
)
docstring_google_tf_adam_str = "\n".join(docstring_google_tf_adam)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/optimizer_v2/adadelta.py#L33-L77
docstring_google_tf_adadelta = (
    "Optimizer that implements the Adadelta algorithm.",
    "",
    "Adadelta optimization is a stochastic gradient descent method that is based on",
    "adaptive learning rate per dimension to address two drawbacks:",
    "",
    "- The continual decay of learning rates throughout training",
    "- The need for a manually selected global learning rate",
    "",
    "Adadelta is a more robust extension of Adagrad that adapts learning rates",
    "based on a moving window of gradient updates, instead of accumulating all",
    "past gradients. This way, Adadelta continues learning even when many updates",
    "have been done. Compared to Adagrad, in the original version of Adadelta you",
    "don't have to set an initial learning rate. In this version, initial",
    "learning rate can be set, as in most other Keras optimizers.",
    "",
    'According to section 4.3 ("Effective Learning rates"), near the end of',
    "training step sizes converge to 1 which is effectively a high learning",
    "rate which would cause divergence. This occurs only near the end of the",
    "training as gradients and step sizes are small, and the epsilon constant",
    "in the numerator and denominator dominate past gradients and parameter",
    "updates which converge the learning rate to 1.",
    "",
    'According to section 4.4("Speech Data"),where a large neural network with',
    "4 hidden layers was trained on a corpus of US English data, ADADELTA was",
    "used with 100 network replicas.The epsilon used is 1e-6 with rho=0.95",
    "which converged faster than ADAGRAD, by the following construction:",
    "def __init__(self, lr=1.0, rho=0.95, epsilon=1e-6, decay=0., **kwargs):",
    "",
    "Args:",
    "learning_rate: A `Tensor`, floating point value, or a schedule that is a",
    "  `tf.keras.optimizers.schedules.LearningRateSchedule`. The learning rate.",
    "  To match the exact form in the original paper use 1.0.",
    "rho: A `Tensor` or a floating point value. The decay rate.",
    "epsilon: A `Tensor` or a floating point value.  A constant epsilon used",
    "         to better conditioning the grad update.",
    "name: Optional name prefix for the operations created when applying",
    '  gradients.  Defaults to `"Adadelta"`.',
    "**kwargs: Keyword arguments. Allowed to be one of",
    '  `"clipnorm"` or `"clipvalue"`.',
    '  `"clipnorm"` (float) clips gradients by norm; `"clipvalue"` (float) clips',
    "  gradients by value.",
    "",
    "Reference:",
    "- [Zeiler, 2012](http://arxiv.org/abs/1212.5701)",
)
docstring_google_tf_adadelta_str = "\n".join(docstring_google_tf_adadelta)

# https://github.com/tensorflow/tensorflow/blob/7ad2723/tensorflow/python/keras/callbacks.py#L2649-L2699
docstring_google_tf_lambda_callback = (
    "Callback for creating simple, custom callbacks on-the-fly.",
    "",
    "  This callback is constructed with anonymous functions that will be called",
    "  at the appropriate time. Note that the callbacks expects positional",
    "  arguments, as:",
    "",
    "  - `on_epoch_begin` and `on_epoch_end` expect two positional arguments:",
    "    `epoch`, `logs`",
    "  - `on_batch_begin` and `on_batch_end` expect two positional arguments:",
    "    `batch`, `logs`",
    "  - `on_train_begin` and `on_train_end` expect one positional argument:",
    "    `logs`",
    "",
    "  Args:",
    "      on_epoch_begin: called at the beginning of every epoch.",
    "      on_epoch_end: called at the end of every epoch.",
    "      on_batch_begin: called at the beginning of every batch.",
    "      on_batch_end: called at the end of every batch.",
    "      on_train_begin: called at the beginning of model training.",
    "      on_train_end: called at the end of model training.",
    "",
    "  Example:",
    "",
    "  ```python",
    "  # Print the batch number at the beginning of every batch.",
    "  batch_print_callback = LambdaCallback(",
    "      on_batch_begin=lambda batch,logs: print(batch))",
    "",
    "  # Stream the epoch loss to a file in JSON format. The file content",
    "  # is not well-formed JSON but rather has a JSON object per line.",
    "  import json",
    "  json_log = open('loss_log.json', mode='wt', buffering=1)",
    "  json_logging_callback = LambdaCallback(",
    "      on_epoch_end=lambda epoch, logs: json_log.write(",
    "          json.dumps({'epoch': epoch, 'loss': logs['loss']}) + '\\n'),",
    "      on_train_end=lambda logs: json_log.close()",
    "  )",
    "",
    "  # Terminate some processes after having finished model training.",
    "  processes = ...",
    "  cleanup_callback = LambdaCallback(",
    "      on_train_end=lambda logs: [",
    "          p.terminate() for p in processes if p.is_alive()])",
    "",
    "  model.fit(...,",
    "            callbacks=[batch_print_callback,",
    "                       json_logging_callback,",
    "                       cleanup_callback])",
    "  ```",
)
docstring_google_tf_lambda_callback_str = "\n".join(docstring_google_tf_lambda_callback)

# ```python
# import ast
# import inspect
# from tensorflow.python.ops.losses.losses_impl import mean_squared_error
#
# ast.get_docstring(ast.parse(inspect.getsource(mean_squared_error)).body[0]).splitlines()
# ```
# https://github.com/tensorflow/tensorflow/blob/v2.7.0/tensorflow/python/ops/losses/losses_impl.py#L624-L771
docstring_google_tf_mean_squared_error_header_tuple = (
    "Adds a Sum-of-Squares loss to the training procedure.",
    "",
    "`weights` acts as a coefficient for the loss. If a scalar is provided, then",
    "the loss is simply scaled by the given value. If `weights` is a tensor of size",
    "`[batch_size]`, then the total loss for each sample of the batch is rescaled",
    "by the corresponding element in the `weights` vector. If the shape of",
    "`weights` matches the shape of `predictions`, then the loss of each",
    "measurable element of `predictions` is scaled by the corresponding value of",
    "`weights`.",
    "",
)
docstring_google_tf_mean_squared_error_args_tuple = (
    "Args:",
    "  labels: The ground truth output tensor, same dimensions as 'predictions'.",
    "  predictions: The predicted outputs.",
    "  weights: Optional `Tensor` whose rank is either 0, or the same rank as",
    "    `labels`, and must be broadcastable to `labels` (i.e., all dimensions must",
    "    be either `1`, or the same as the corresponding `losses` dimension).",
    "  scope: The scope for the operations performed in computing the loss.",
    "  loss_collection: collection to which the loss will be added.",
    "  reduction: Type of reduction to apply to loss.",
    "",
    "Returns:",
    "  Weighted loss float `Tensor`. If `reduction` is `NONE`, this has the same",
    "  shape as `labels`; otherwise, it is scalar.",
    "",
)
docstring_google_tf_mean_squared_error_footer_tuple = (
    "Raises:",
    "  ValueError: If the shape of `predictions` doesn't match that of `labels` or",
    "    if the shape of `weights` is invalid.  Also if `labels` or `predictions`",
    "    is None.",
    "",
    "@compatibility(TF2)",
    "",
    "`tf.compat.v1.losses.mean_squared_error` is mostly compatible with eager",
    "execution and `tf.function`. But, the `loss_collection` argument is",
    "ignored when executing eagerly and no loss will be written to the loss",
    "collections. You will need to either hold on to the return value manually",
    "or rely on `tf.keras.Model` loss tracking.",
    "",
    "",
    "To switch to native TF2 style, instantiate the",
    " `tf.keras.losses.MeanSquaredError` class and call the object instead.",
    "",
    "",
    "#### Structural Mapping to Native TF2",
    "",
    "Before:",
    "",
    "```python",
    "loss = tf.compat.v1.losses.mean_squared_error(",
    "  labels=labels,",
    "  predictions=predictions,",
    "  weights=weights,",
    "  reduction=reduction)",
    "```",
    "",
    "After:",
    "",
    "```python",
    "loss_fn = tf.keras.losses.MeanSquaredError(",
    "  reduction=reduction)",
    "loss = loss_fn(",
    "  y_true=labels,",
    "  y_pred=predictions,",
    "  sample_weight=weights)",
    "```",
    "",
    "#### How to Map Arguments",
    "",
    "| TF1 Arg Name          | TF2 Arg Name     | Note                       |",
    "| :-------------------- | :--------------- | :------------------------- |",
    "| `labels`              | `y_true`         | In `__call__()` method     |",
    "| `predictions`         | `y_pred`         | In `__call__()` method     |",
    "| `weights`             | `sample_weight`  | In `__call__()` method.    |",
    ": : : The shape requirements for `sample_weight` is different from      :",
    ": : : `weights`. Please check the [argument definition][api_docs] for   :",
    ": : : details.                                                          :",
    "| `scope`               | Not supported    | -                          |",
    "| `loss_collection`     | Not supported    | Losses should be tracked   |",
    ": : : explicitly or with Keras APIs, for example, [add_loss][add_loss], :",
    ": : : instead of via collections                                        :",
    "| `reduction`           | `reduction`      | In constructor. Value of   |",
    ": : : `tf.compat.v1.losses.Reduction.SUM_OVER_BATCH_SIZE`,              :",
    ": : : `tf.compat.v1.losses.Reduction.SUM`,                              :",
    ": : : `tf.compat.v1.losses.Reduction.NONE` in                           :",
    ": : : `tf.compat.v1.losses.softmax_cross_entropy` correspond to         :",
    ": : : `tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE`,                  :",
    ": : : `tf.keras.losses.Reduction.SUM`,                                  :",
    ": : : `tf.keras.losses.Reduction.NONE`, respectively. If you            :",
    ": : : used other value for `reduction`, including the default value     :",
    ": : :  `tf.compat.v1.losses.Reduction.SUM_BY_NONZERO_WEIGHTS`, there is :",
    ": : : no directly corresponding value. Please modify the loss           :",
    ": : : implementation manually.                                          :",
    "",
    "[add_loss]:https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#add_loss",
    "[api_docs]:https://www.tensorflow.org/api_docs/python/tf/keras/losses/MeanSquaredError#__call__",
    "",
    "",
    "#### Before & After Usage Example",
    "",
    "Before:",
    "",
    ">>> y_true = [1, 2, 3]",
    ">>> y_pred = [1, 3, 5]",
    ">>> weights = [0, 1, 0.25]",
    ">>> # samples with zero-weight are excluded from calculation when `reduction`",
    ">>> # argument is set to default value `Reduction.SUM_BY_NONZERO_WEIGHTS`",
    ">>> tf.compat.v1.losses.mean_squared_error(",
    "...    labels=y_true,",
    "...    predictions=y_pred,",
    "...    weights=weights).numpy()",
    "1.0",
    "",
    ">>> tf.compat.v1.losses.mean_squared_error(",
    "...    labels=y_true,",
    "...    predictions=y_pred,",
    "...    weights=weights,",
    "...    reduction=tf.compat.v1.losses.Reduction.SUM_OVER_BATCH_SIZE).numpy()",
    "0.66667",
    "",
    "After:",
    "",
    ">>> y_true = [[1.0], [2.0], [3.0]]",
    ">>> y_pred = [[1.0], [3.0], [5.0]]",
    ">>> weights = [1, 1, 0.25]",
    ">>> mse = tf.keras.losses.MeanSquaredError(",
    "...    reduction=tf.keras.losses.Reduction.SUM_OVER_BATCH_SIZE)",
    ">>> mse(y_true=y_true, y_pred=y_pred, sample_weight=weights).numpy()",
    "0.66667",
    "",
    "@end_compatibility",
)
docstring_google_tf_mean_squared_error_str = "\n".join(
    chain.from_iterable(
        (
            docstring_google_tf_mean_squared_error_header_tuple,
            docstring_google_tf_mean_squared_error_args_tuple,
            docstring_google_tf_mean_squared_error_footer_tuple,
        )
    )
)

docstring_google_pytorch_lbfgs = (
    "",
    "Implements L-BFGS algorithm, heavily inspired by `minFunc",
    "    <https://www.cs.ubc.ca/~schmidtm/Software/minFunc.html>`.",
    "",
    "    Args:",
    "        lr (float): learning rate (default: 1)",
    "        max_iter (int): maximal number of iterations per optimization step",
    "            (default: 20)",
    "        max_eval (int): maximal number of function evaluations per optimization",
    "            step (default: max_iter * 1.25).",
    "        tolerance_grad (float): termination tolerance on first order optimality",
    "            (default: 1e-5).",
    "        tolerance_change (float): termination tolerance on function",
    "            value/parameter changes (default: 1e-9).",
    "        history_size (int): update history size (default: 100).",
    "        line_search_fn (str): either 'strong_wolfe' or None (default: None).",
)
docstring_google_pytorch_lbfgs_str = "\n".join(docstring_google_pytorch_lbfgs)

docstring_google_str = """{docstring_header_str}
Args:
  dataset_name (str): name of dataset. Defaults to "mnist"
  tfds_dir (str): directory to look for models in. Defaults to "~/tensorflow_datasets"
  K (Literal['np', 'tf']): backend engine, e.g., `np` or `tf`. Defaults to "np"
  as_numpy (Optional[bool]): Convert to numpy ndarrays
  data_loader_kwargs (Optional[dict]): pass this as arguments to data_loader function

Returns:
  Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]:
   Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
""".format(
    docstring_header_str=docstring_header_str
)

docstring_no_default_doc_str = """
{header_doc_str}
:param dataset_name: name of dataset.
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in.
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`.
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:return: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
""".format(
    header_doc_str=docstring_header_str
)

docstring_no_default_doc_wrapped_str = docstring_no_default_doc_str.replace(
    " np.ndarray]]```", "\n{tab}np.ndarray]]```".format(tab=tab)
)

_docstring_no_default_str = """
{header_doc_str}\n
:param dataset_name: name of dataset.
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in.
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`.
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:return: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
"""

docstring_no_default_str = _docstring_no_default_str.format(
    header_doc_str=docstring_header_str
)

docstring_no_default_no_nl_str = _docstring_no_default_str.format(
    header_doc_str=docstring_header_no_nl_str
)

docstring_numpydoc_only_doc_str = """
{header_doc_str}
""".format(
    header_doc_str=docstring_header_str
)

docstring_numpydoc_only_params_str = """
Parameters
----------
dataset_name : str
    name of dataset. Defaults to "mnist"
tfds_dir : str
    directory to look for models in. Defaults to "~/tensorflow_datasets"
K : Literal['np', 'tf']
    backend engine, e.g., `np` or `tf`. Defaults to "np"
as_numpy : Optional[bool]
    Convert to numpy ndarrays
data_loader_kwargs : Optional[dict]
    pass this as arguments to data_loader function
"""

docstring_numpydoc_only_returns_str = """
Returns
-------
Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]
    Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
"""

docstring_numpydoc_str = """{docstring_header_str}
Parameters
----------
dataset_name : str
    name of dataset. Defaults to "mnist"
tfds_dir : str
    directory to look for models in. Defaults to "~/tensorflow_datasets"
K : Literal['np', 'tf']
    backend engine, e.g., `np` or `tf`. Defaults to "np"
as_numpy : Optional[bool]
    Convert to numpy ndarrays
data_loader_kwargs : Optional[dict]
    pass this as arguments to data_loader function

{docstring_numpydoc_only_returns_str}""".format(
    docstring_header_str=docstring_header_str,
    docstring_numpydoc_only_returns_str=docstring_numpydoc_only_returns_str.lstrip(
        "\n"
    ),
)

docstring_only_return_type_str = """
Some comment

:param dataset_name: Example: foo

:return: Train and tests dataset splits.
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
"""

_docstring_str = """
{header_doc_str}\n
:param dataset_name: name of dataset. Defaults to "mnist"
:type dataset_name: ```str```

:param tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"
:type tfds_dir: ```str```

:param K: backend engine, e.g., `np` or `tf`. Defaults to "np"
:type K: ```Literal['np', 'tf']```

:param as_numpy: Convert to numpy ndarrays
:type as_numpy: ```Optional[bool]```

:param data_loader_kwargs: pass this as arguments to data_loader function
:type data_loader_kwargs: ```Optional[dict]```

:return: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
:rtype: ```Union[Tuple[tf.data.Dataset, tf.data.Dataset], Tuple[np.ndarray, np.ndarray]]```
"""

docstring_str = _docstring_str.format(header_doc_str=docstring_header_str)

docstring_no_nl_str = _docstring_str.format(header_doc_str=docstring_header_no_nl_str)

docstring_no_nl_no_none_str = docstring_no_nl_str.replace(" Defaults to None", "")

docstring_wrapped_str = docstring_str.replace(
    " np.ndarray]]```", "\n{tab}np.ndarray]]```".format(tab=tab)
)

docstring_no_type_str = """
{header_doc_str}

:param dataset_name: name of dataset. Defaults to "mnist"

:param tfds_dir: directory to look for models in. Defaults to "~/tensorflow_datasets"

:param K: backend engine, e.g., `np` or `tf`. Defaults to "np"

:param as_numpy: Convert to numpy ndarrays

:param data_loader_kwargs: pass this as arguments to data_loader function

:return: Train and tests dataset splits. Defaults to (np.empty(0), np.empty(0))
""".format(
    header_doc_str=docstring_header_str
)

docstring_no_type_no_default_tpl_str = """
{header_doc_str}
:param dataset_name: name of dataset.

:param tfds_dir: directory to look for models in.

:param K: backend engine, e.g., `np` or `tf`.

:param as_numpy: Convert to numpy ndarrays

:param data_loader_kwargs: pass this as arguments to data_loader function

:return: Train and tests dataset splits.
"""

docstring_no_type_no_default_str = docstring_no_type_no_default_tpl_str.format(
    header_doc_str=docstring_header_str
)

docstring_repr_str = (
    indent(
        "\n".join(
            (
                "",
                "Emit a string representation of the current instance",
                "",
                ":return: String representation of instance",
                ":rtype: ```str```",
                "",
            )
        ),
        tab * 2,
        identity,
    )
    + tab * 2
)

# docstring_repr_google_str = emit.docstring(parse.docstring(docstring_repr_str), docstring_format="google")
docstring_repr_google_str = (
    "\nEmit a string representation of the current instance\n\n\n\n\n"
    "Returns:\n"
    "  str:\n"
    "   String representation of instance\n"
)

docstring_reduction_v2_str = (
    "Types of loss reduction."
    "\n"
    "\n"
    "  Contains the following values:"
    "\n"
    "\n"
    "  * `AUTO`: Indicates that the reduction option will be determined by the usage\n"
    "     context. For almost all cases this defaults to `SUM_OVER_BATCH_SIZE`. When\n"
    "     used with `tf.distribute.Strategy`, outside of built-in training loops such\n"
    "     as `tf.keras` `compile` and `fit`, we expect reduction value to be\n"
    "     `SUM` or `NONE`. Using `AUTO` in that case will raise an error.\n"
    "  * `NONE`: No **additional** reduction is applied to the output of the wrapped\n"
    "     loss function. When non-scalar losses are returned to Keras functions like\n"
    "     `fit`/`evaluate`, the unreduced vector loss is passed to the optimizer\n"
    "     but the reported loss will be a scalar value.\n"
    "\n"
    "     Caution: **Verify the shape of the outputs when using** `Reduction.NONE`.\n"
    "     The builtin loss functions wrapped by the loss classes reduce\n"
    "     one dimension (`axis=-1`, or `axis` if specified by loss function).\n"
    "     `Reduction.NONE` just means that no **additional** reduction is applied by\n"
    "     the class wrapper. For categorical losses with an example input shape of\n"
    "     `[batch, W, H, n_classes]` the `n_classes` dimension is reduced. For\n"
    "     pointwise losses your must include a dummy axis so that `[batch, W, H, 1]`\n"
    "     is reduced to `[batch, W, H]`. Without the dummy axis `[batch, W, H]`\n"
    "     will be incorrectly reduced to `[batch, W]`.\n"
    "\n"
    "  * `SUM`: Scalar sum of weighted losses.\n"
    "  * `SUM_OVER_BATCH_SIZE`: Scalar `SUM` divided by number of elements in losses.\n"
    "     This reduction type is not supported when used with\n"
    "     `tf.distribute.Strategy` outside of built-in training loops like `tf.keras`\n"
    "     `compile`/`fit`.\n"
    "\n"
    "     You can implement 'SUM_OVER_BATCH_SIZE' using global batch size like:\n"
    "     ```\n"
    "     with strategy.scope():\n"
    "       loss_obj = tf.keras.losses.CategoricalCrossentropy(\n"
    "           reduction=tf.keras.losses.Reduction.NONE)\n"
    "       ....\n"
    "       loss = tf.reduce_sum(loss_obj(labels, predictions)) *\n"
    "           (1. / global_batch_size)\n"
    "     ```\n"
    "\n"
    "  Please see the [custom training guide](\n"
    "  https://www.tensorflow.org/tutorials/distribute/custom_training) for more\n"
    "  details on this.\n"
    "  "
)

docstring_keras_rmsprop_class_str = (
    "Optimizer that implements the RMSprop algorithm.\n"
    "\n"
    "  The gist of RMSprop is to:\n"
    "\n  - Maintain a moving (discounted) average of the square of gradients\n"
    "  - Divide the gradient by the root of this average\n"
    "\n"
    "  This implementation of RMSprop uses plain momentum, not Nesterov momentum.\n"
    "\n"
    "  The centered version additionally maintains a moving average of the\n"
    "  gradients, and uses that average to estimate the variance.\n"
    "\n"
    "  Args:\n"
    "    learning_rate: A `Tensor`, floating point value, or a schedule that is a\n"
    "      `tf.keras.optimizers.schedules.LearningRateSchedule`, or a callable\n"
    "      that takes no arguments and returns the actual value to use. The\n"
    "      learning rate. Defaults to 0.001.\n"
    "    rho: Discounting factor for the history/coming gradient. Defaults to 0.9.\n"
    "    momentum: A scalar or a scalar `Tensor`. Defaults to 0.0.\n"
    "    epsilon: A small constant for numerical stability. This epsilon is\n"
    '      "epsilon hat" in the Kingma and Ba paper (in the formula just before\n'
    "      Section 2.1), not the epsilon in Algorithm 1 of the paper. Defaults to\n"
    "      1e-7.\n"
    "    centered: Boolean. If `True`, gradients are normalized by the estimated\n"
    "      variance of the gradient; if False, by the uncentered second moment.\n"
    "      Setting this to `True` may help with training, but is slightly more\n"
    "      expensive in terms of computation and memory. Defaults to `False`.\n"
    "    name: Optional name prefix for the operations created when applying\n"
    '      gradients. Defaults to `"RMSprop"`.\n'
    "    **kwargs: Keyword arguments. Allowed to be one of\n"
    '      `"clipnorm"` or `"clipvalue"`.\n'
    '      `"clipnorm"` (float) clips gradients by norm; `"clipvalue"` (float) clips\n'
    "      gradients by value.\n"
    "\n"
    "  Note that in the dense implementation of this algorithm, variables and their\n"
    "  corresponding accumulators (momentum, gradient moving average, square\n"
    "  gradient moving average) will be updated even if the gradient is zero\n"
    "  (i.e. accumulators will decay, momentum will be applied). The sparse\n"
    "  implementation (used when the gradient is an `IndexedSlices` object,\n"
    "  typically because of `tf.gather` or an embedding lookup in the forward pass)\n"
    "  will not update variable slices or their accumulators unless those slices\n"
    '  were used in the forward pass (nor is there an "eventual" correction to\n'
    "  account for these omitted updates). This leads to more efficient updates for\n"
    "  large embedding lookup tables (where most of the slices are not accessed in\n"
    "  a particular graph execution), but differs from the published algorithm.\n"
    "\n"
    "  Usage:\n"
    "\n"
    "  >>> opt = tf.keras.optimizers.RMSprop(learning_rate=0.1)\n"
    "  >>> var1 = tf.Variable(10.0)\n"
    "  >>> loss = lambda: (var1 ** 2) / 2.0    # d(loss) / d(var1) = var1\n"
    "  >>> step_count = opt.minimize(loss, [var1]).numpy()\n"
    "  >>> var1.numpy()\n"
    "  9.683772\n"
    "\n"
    "  Reference:\n"
    "    - [Hinton, 2012](\n"
    "      http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)\n"
    "  "
)

docstring_keras_rmsprop_method_str = (
    "Construct a new RMSprop optimizer.\n"
    "\n"
    "    Args:\n"
    "      learning_rate: A `Tensor`, floating point value, or a schedule that is a\n"
    "        `tf.keras.optimizers.schedules.LearningRateSchedule`, or a callable\n"
    "        that takes no arguments and returns the actual value to use. The\n"
    "        learning rate. Defaults to 0.001.\n"
    "      rho: Discounting factor for the history/coming gradient. Defaults to 0.9.\n"
    "      momentum: A scalar or a scalar `Tensor`. Defaults to 0.0.\n"
    "      epsilon: A small constant for numerical stability. This epsilon is\n"
    '        "epsilon hat" in the Kingma and Ba paper (in the formula just before\n'
    "        Section 2.1), not the epsilon in Algorithm 1 of the paper. Defaults to\n"
    "        1e-7.\n"
    "      centered: Boolean. If `True`, gradients are normalized by the estimated\n"
    "        variance of the gradient; if False, by the uncentered second moment.\n"
    "        Setting this to `True` may help with training, but is slightly more\n"
    "        expensive in terms of computation and memory. Defaults to `False`.\n"
    "      name: Optional name prefix for the operations created when applying\n"
    '        gradients. Defaults to "RMSprop".\n'
    "      **kwargs: keyword arguments. Allowed to be {`clipnorm`, `clipvalue`, `lr`,\n"
    "        `decay`}. `clipnorm` is clip gradients by norm; `clipvalue` is clip\n"
    "        gradients by value, `decay` is included for backward compatibility to\n"
    "        allow time inverse decay of learning rate. `lr` is included for backward\n"
    "        compatibility, recommended to use `learning_rate` instead.\n"
    "\n"
    "    @compatibility(eager)\n"
    "    When eager execution is enabled, `learning_rate`, `decay`, `momentum`, and\n"
    "    `epsilon` can each be a callable that takes no arguments and returns the\n"
    "    actual value to use. This can be useful for changing these values across\n"
    "    different invocations of optimizer functions.\n"
    "    @end_compatibility\n"
    "    "
)

docstring_google_tf_ops_losses__safe_mean_str = (
    "Computes a safe mean of the losses.\n"
    "\n"
    "  Args:\n"
    "    losses: `Tensor` whose elements contain individual loss measurements.\n"
    "    num_present: The number of measurable elements in `losses`.\n"
    "\n"
    "  Returns:\n"
    "    A scalar representing the mean of `losses`. If `num_present` is zero,\n"
    "      then zero is returned.\n"
    "  "
)

docstring_sum_tuple = (
    ":type a: ```int```",
    "",
    ":type b: ```int```",
    "",
    ":rtype: ```int```",
    "",
)

__all__ = [
    "docstring_extra_colons_str",
    "docstring_google_str",
    "docstring_google_tf_adadelta_str",
    "docstring_google_tf_adam_str",
    "docstring_google_tf_lambda_callback_str",
    "docstring_google_tf_mean_squared_error_args_tuple",
    "docstring_google_tf_mean_squared_error_footer_tuple",
    "docstring_google_tf_mean_squared_error_header_tuple",
    "docstring_google_tf_mean_squared_error_str",
    "docstring_google_tf_ops_losses__safe_mean_str",
    "docstring_google_tf_squared_hinge_str",
    "docstring_header_and_return_no_nl_str",
    "docstring_header_and_return_str",
    "docstring_header_and_return_two_nl_str",
    "docstring_header_no_nl_str",
    "docstring_header_str",
    "docstring_keras_rmsprop_class_str",
    "docstring_keras_rmsprop_method_str",
    "docstring_no_default_doc_str",
    "docstring_no_default_no_nl_str",
    "docstring_no_default_str",
    "docstring_no_nl_no_none_str",
    "docstring_no_nl_str",
    "docstring_numpydoc_only_doc_str",
    "docstring_numpydoc_only_params_str",
    "docstring_numpydoc_only_returns_str",
    "docstring_numpydoc_str",
    "docstring_only_return_type_str",
    "docstring_reduction_v2_str",
    "docstring_repr_google_str",
    "docstring_repr_str",
    "docstring_str",
    "docstring_sum_tuple",
]



"""
Concrete Syntax Tree utility functions
"""

from collections import OrderedDict, deque, namedtuple
from dataclasses import make_dataclass
from functools import partial, wraps
from keyword import kwlist
from typing import Optional

from cdd.shared.pure_utils import balanced_parentheses, is_triple_quoted, tab

kwset = frozenset(kwlist)
_basic_cst_attributes = "line_no_start", "line_no_end", "value"
UnchangingLine = namedtuple("UnchangingLine", _basic_cst_attributes)
Assignment = namedtuple("Assignment", _basic_cst_attributes)
AnnAssignment = namedtuple("AnnAssignment", _basic_cst_attributes)
AugAssignment = namedtuple("AugAssignment", _basic_cst_attributes)
# Same for `AsyncFunctionDefinition`
FunctionDefinitionStart = namedtuple(
    "FunctionDefinitionStart", (*_basic_cst_attributes, "name")
)
ClassDefinitionStart = namedtuple(
    "ClassDefinitionStart", (*_basic_cst_attributes, "name")
)

# compound_stmt needs all of them to differentiate which indentor is scope-creating
IfStatement = namedtuple("IfStatement", _basic_cst_attributes)
ElifStatement = namedtuple("ElifStatement", _basic_cst_attributes)
ElseStatement = namedtuple("ElseStatement", _basic_cst_attributes)
WithStatement = namedtuple("WithStatement", _basic_cst_attributes)
ForStatement = namedtuple("ForStatement", _basic_cst_attributes)
WhileStatement = namedtuple("WhileStatement", _basic_cst_attributes)
MatchStatement = namedtuple("MatchStatement", _basic_cst_attributes)
CaseStatement = namedtuple("CaseStatement", _basic_cst_attributes)
# | function_def
# | if_stmt
# | class_def
# | with_stmt
# | for_stmt
# | try_stmt
# | while_stmt
# | match_stmt

CommentStatement = namedtuple("CommentStatement", _basic_cst_attributes)
PassStatement = namedtuple("PassStatement", _basic_cst_attributes)
DelStatement = namedtuple("DelStatement", _basic_cst_attributes)
YieldStatement = namedtuple("YieldStatement", _basic_cst_attributes)
BreakStatement = namedtuple("BreakStatement", _basic_cst_attributes)
ContinueStatement = namedtuple("ContinueStatement", _basic_cst_attributes)
GlobalStatement = namedtuple("GlobalStatement", _basic_cst_attributes)
NonlocalStatement = namedtuple("NonlocalStatement", _basic_cst_attributes)
ReturnStatement = namedtuple("ReturnStatement", _basic_cst_attributes)
RaiseStatement = namedtuple("RaiseStatement", _basic_cst_attributes)
TryStatement = namedtuple("TryStatement", _basic_cst_attributes)
CatchStatement = namedtuple("TryStatement", _basic_cst_attributes)
ExceptStatement = namedtuple("ExceptStatement", _basic_cst_attributes)
FinallyStatement = namedtuple("FinallyStatement", _basic_cst_attributes)
FromStatement = namedtuple("FromStatement", _basic_cst_attributes)
ImportStatement = namedtuple("ImportStatement", _basic_cst_attributes)
TrueStatement = namedtuple("TrueStatement", _basic_cst_attributes)
FalseStatement = namedtuple("FalseStatement", _basic_cst_attributes)
NoneStatement = namedtuple("NoneStatement", _basic_cst_attributes)
ExprStatement = namedtuple("ExprStatement", _basic_cst_attributes)
SetExprStatement = namedtuple("SetExprStatement", _basic_cst_attributes)
DictExprStatement = namedtuple("DictExprStatement", _basic_cst_attributes)
GenExprStatement = namedtuple("GenExprStatement", _basic_cst_attributes)
ListCompStatement = namedtuple("ListCompStatement", _basic_cst_attributes)
CallStatement = namedtuple("CallStatement", _basic_cst_attributes)

contains2statement = OrderedDict(
    (
        ("#", CommentStatement),
        ("pass", PassStatement),
        ("del", DelStatement),
        ("yield", YieldStatement),
        ("break", BreakStatement),
        ("continue", ContinueStatement),
        ("global", GlobalStatement),
        ("nonlocal", NonlocalStatement),
        ("return", ReturnStatement),
        ("raise", RaiseStatement),
        ("except", ExceptStatement),
        ("finally", FinallyStatement),
        ("try", TryStatement),
        ("from", FromStatement),
        ("import", ImportStatement),
        ("if", IfStatement),
        ("elif", ElifStatement),
        ("else:", ElseStatement),
        ("with", WithStatement),
        ("for", ForStatement),
        ("while", WhileStatement),
        ("match", MatchStatement),
        ("case", CaseStatement),
        ("True", TrueStatement),
        ("False", FalseStatement),
        ("None", NoneStatement),
    )
)
math_operators = frozenset(
    (
        "@",
        "/",
        "//",
        "*",
        "**",
        "+",
        "-",
        "%",
        "&",
        "|",
        "<<",
        ">>",
        "<",
        ">",
        "==",
        ">=",
        "<=",
        "^",
    )
)

augassign = frozenset(
    ("+=" "-=", "*=", "@=", "/=", "%=", "&=", "|=", "^=", "<<=", ">>=", "**=", "//=")
)
# ':=' could be an augassign I guess
multicontains2statement = (
    (frozenset((":", "=")), AnnAssignment),  # TODO: support `class F: foo: int` also…
    (frozenset(("=",)), Assignment),
)

ast2cst = {
    "Expr": ExprStatement,
    "AnnAssign": AnnAssignment,
    "Assign": Assignment,
    "Constant": ExprStatement,
    "Str": ExprStatement,
    "Num": ExprStatement,
    "ClassDef": ClassDefinitionStart,
    "FunctionDef": FunctionDefinitionStart,
    "AsyncFunctionDef": FunctionDefinitionStart,
    # TODO: Full spec
}

TripleQuoted = make_dataclass(
    "TripleQuoted",
    [
        ("is_double_q", Optional[bool]),
        ("is_docstr", Optional[bool]),
        ("line_no_start", int),
        ("line_no_end", int),
        ("value", str),
    ],
    namespace={},
)


def get_construct_name(words):
    """
    Find the construct name, currently works for:
    - AsyncFunctionDef
    - FunctionDef
    - ClassDef

    :param words: Tuple of words (no whitespace)
    :type words: ```Tuple[str]```

    :return: Name of construct if found else None
    :rtype: ```Optional[str]```
    """
    for idx, word in enumerate(words):
        if len(words) > idx + 1:
            if word == "def":
                return words[idx + 1][: words[idx + 1].find("(")]
            elif word == "class":
                end_idx = (
                    lambda _end_idx: words[idx + 1].find(":")
                    if _end_idx == -1
                    else _end_idx
                )(words[idx + 1].find("("))
                return words[idx + 1][:end_idx]


def cst_scanner(source):
    """
    Reduce source code into chunks useful for parsing.
    These chunks include newlines and the array is one dimensional.

    :param source: Python source code
    :type source: ```str```

    :return: List of scanned source code
    :rtype: ```List[str]```
    """
    scanned, stack = [], []
    for idx, ch in enumerate(source):
        if ch == "\n":  # in frozenset(("\n", ":", ";", '"""', "'''", '#')):
            cst_scan(scanned, stack)
        stack.append(ch)
    cst_scan(scanned, stack)
    if stack:
        scanned.append("".join(stack))
    return scanned


def cst_scan(scanned, stack):
    """
    Checks if what has been scanned (stack) is ready for the `scanned` array
    Add then clear stack if ready else do nothing with both

    :param scanned: List of statements observed
    :type scanned: ```List[str]```

    :param stack: List of characters observed
    :type stack: ```List[str]```
    """
    statement = "".join(stack)
    statement_stripped = statement.strip()

    is_comment = statement_stripped.startswith("#")
    if statement_stripped.endswith("\\"):
        has_triple_quotes = is_other_statement = False
    else:
        has_triple_quotes = is_triple_quoted(statement_stripped)
        is_other_statement = all(
            (
                statement_stripped,
                balanced_parentheses(statement_stripped),
                not statement_stripped.startswith("@")
                or statement_stripped.endswith(":"),
                not statement_stripped.startswith("'''"),
                not statement_stripped.startswith('"""'),
            )
        )

    statement_found = is_comment or has_triple_quotes or is_other_statement
    if statement_found:
        if is_comment:
            scanned.append(statement)
            stack.clear()
        elif is_other_statement:
            expression = []

            def add_and_clear(the_expression_str, expr, scanned_tokens, the_stack):
                """
                Add the found statement and clear the stacks

                :param the_expression_str: Expression string
                :type the_expression_str: ```str```

                :param expr: The expression
                :type expr: ```List[str]```

                :param scanned_tokens: Scanned tokens
                :type scanned_tokens: ```List[str]```

                :param the_stack: The current stack
                :type the_stack: ```List[str]```

                """
                scanned_tokens.append(the_expression_str)
                expr.clear()
                the_stack.clear()

            expression_str = ""
            for idx, ch in enumerate(statement):
                expression.append(ch)
                expression_str = "".join(expression)
                expression_stripped = expression_str.strip()

                if (
                    is_triple_quoted(expression_stripped)
                    or expression_stripped.startswith("#")
                    and expression_str.endswith("\n")
                ):
                    add_and_clear(expression_str, expression, scanned, stack)
                    # Single comment should be picked up
                elif balanced_parentheses(expression_stripped):
                    if (
                        expression_str.endswith("\n")
                        and not expression_stripped.endswith("\\")
                        and (
                            not expression_stripped.endswith(":")
                            or "class" not in expression_stripped
                            and "def" not in expression_stripped
                        )
                        and not expression_str.isspace()
                        and not expression_stripped.startswith("@")
                    ):
                        add_and_clear(expression_str, expression, scanned, stack)
                    else:
                        words = tuple(
                            filter(None, map(str.strip, expression_stripped.split(" ")))
                        )
                        if "def" in words or "class" in words:
                            if words[-1].endswith(":") and balanced_parentheses(
                                expression_stripped
                            ):
                                add_and_clear(
                                    expression_str, expression, scanned, stack
                                )
                        # elif ";"
            if expression:
                add_and_clear(expression_str, expression, scanned, stack)
            # Longest matching pattern should be parsed out but this does shortest^
        else:
            scanned.append(statement)
            stack.clear()


def cst_parser(scanned):
    """
    Checks if what has been scanned (stack) is ready for the `scanned` array
    Add then clear stack if ready else do nothing with both

    :param scanned: List of statements observed
    :type scanned: ```List[str]```

    :return: Parse of scanned statements. One dimensional.
    :rtype: ```Tuple[NamedTuple]```
    """
    state = {
        "acc": 1,
        "prev_node": UnchangingLine(line_no_start=None, line_no_end=None, value=""),
        "parsed": [],
    }
    # return accumulate(scanned, partial(cst_parse_one_node, state=state), initial=[])
    deque(map(partial(cst_parse_one_node, state=state), scanned), maxlen=0)
    return tuple(state["parsed"])


def infer_cst_type(statement_stripped, words):
    """
    Infer the CST type. This is the most important function of the CST parser!

    :param statement_stripped: Original scanned statement minus both ends of whitespace
    :type statement_stripped: ```str```

    :param words: List of whitespace stripped and empty-str removed words from original statement
    :type words: ```List[str]```

    :return: CST type… a NamedTuple with at least ("line_no_start", "line_no_end", "value") attributes
    :rtype: ```NamedTuple```
    """
    if statement_stripped.startswith("["):
        return ListCompStatement
    elif statement_stripped.startswith("{"):
        # Won't work with nested dict inside a SetExpr
        if ":" in statement_stripped:
            return DictExprStatement
        return SetExprStatement
    elif statement_stripped.startswith("("):
        return GenExprStatement

    for word in words:
        if word in contains2statement:
            return contains2statement[word]

    if any(aug_assign in statement_stripped for aug_assign in augassign):
        return AugAssignment

    statement_frozenset = frozenset(statement_stripped)
    for key, constructor in multicontains2statement:
        if statement_frozenset & key == key:
            return constructor

    if any(math_operator in statement_stripped for math_operator in math_operators):
        return ExprStatement

    elif "(" in statement_frozenset:
        return CallStatement

    return UnchangingLine  # or: raise NotImplementedError(repr(statement_stripped))


def set_prev_node(function):
    """
    Store the result of the current function run into the `prev_node` attribute

    :param function: The `cst_parse_one_node` function
    :type function: ```Callable[[str, dict], NamedTuple]```

    :return: Wrapped function
    :rtype: ```Callable[[], Callable[[str, dict], NamedTuple]]```
    """

    @wraps(function)
    def wrapper(statement, state):
        """
        Store the result of parsing one statement into a CST node in the `prev_node` key of the `state` dict

        :param statement: Statement that was scanned from Python source code
        :type statement: ```str```

        :param state: Number of lines between runs in `acc` key; `prev_node`
        :type state: ```dict```

        :return: NamedTuple with at least ("line_no_start", "line_no_end", "value") attributes
        :rtype: ```NamedTuple```
        """
        state["prev_node"] = function(statement, state)
        state["parsed"].append(state["prev_node"])
        return state["prev_node"]

    return wrapper


@set_prev_node
def cst_parse_one_node(statement, state):
    """
    Parse one statement into a CST node

    :param statement: Statement that was scanned from Python source code
    :type statement: ```str```

    :param state: Number of lines between runs in `acc` key; `prev_node`; `parsed`
    :type state: ```dict```

    :return: NamedTuple with at least ("line_no_start", "line_no_end", "value") attributes
    :rtype: ```NamedTuple```
    """
    prev_acc = state["acc"]
    state["acc"] += statement.count("\n")

    statement_stripped = statement.strip()

    words = tuple(filter(None, map(str.strip, statement_stripped.split(" "))))

    common_kwargs = dict(
        line_no_start=prev_acc,
        line_no_end=state["acc"],
        value=statement,
    )

    if len(statement_stripped) > 5:
        is_single_quote = (statement_stripped[:3], statement_stripped[-3:]) == (
            "'''",
            "'''",
        )
        is_double_quote = not is_single_quote and (
            statement_stripped[:3],
            statement_stripped[-3:],
        ) == (
            '"""',
            '"""',
        )
        if is_single_quote or is_double_quote:
            return TripleQuoted(
                **common_kwargs,
                is_double_q=is_double_quote,
                is_docstr=isinstance(
                    state["prev_node"], (ClassDefinitionStart, FunctionDefinitionStart)
                ),
            )

    if words:
        if len(words) > 1:
            name = get_construct_name(words)
            if name is not None:
                return (
                    ClassDefinitionStart
                    if words[0] == "class"
                    else FunctionDefinitionStart
                )(**common_kwargs, name=name)

        return infer_cst_type(statement_stripped, words)(**common_kwargs)

    return UnchangingLine(**common_kwargs)


def reindent_block_with_pass_body(s):
    """
    Reindent block (e.g., function definition) and give it a `pass` body

    :param s: Block defining string
    :type s: ```str```

    :return: Reindented string with `pass` body
    :rtype: ```str```
    """
    return "{block_def} pass".format(
        block_def="\n".join(
            map(
                str.lstrip,
                s.split("\n"),
            )
        ).replace(tab, "", 1)
    )


__all__ = [
    "AnnAssignment",
    "Assignment",
    "AugAssignment",
    "CallStatement",
    "ClassDefinitionStart",
    "CommentStatement",
    "ElifStatement",
    "ElseStatement",
    "ExprStatement",
    "FalseStatement",
    "FromStatement",
    "FunctionDefinitionStart",
    "IfStatement",
    "NoneStatement",
    "PassStatement",
    "ReturnStatement",
    "TripleQuoted",
    "TrueStatement",
    "UnchangingLine",
    "ast2cst",
    "cst_parser",
    "cst_scanner",
    "infer_cst_type",
    "kwset",
    "reindent_block_with_pass_body",
]



"""
Tests `cdd.routes.emit` EMITTERS
"""

from os import path
from unittest import TestCase

from cdd.routes.emit import EMITTERS
from cdd.shared.pure_utils import all_dunder_for_module
from cdd.tests.utils_for_tests import unittest_main


class TestRoutesEmit(TestCase):
    """Tests `cdd.routes.emit`"""

    def test_routes_emit_root(self) -> None:
        """Confirm that route emitter names are up-to-date"""
        self.assertListEqual(
            EMITTERS,
            all_dunder_for_module(
                path.join(
                    path.dirname(path.dirname(path.dirname(__file__))), "routes", "emit"
                ),
                iter(()),
                path_validator=path.isfile,
            ),
        )


unittest_main()



""" Tests for gen_utils """

from unittest import TestCase

from cdd.compound.gen_utils import get_input_mapping_from_path
from cdd.tests.utils_for_tests import unittest_main


def f(s):
    """
    :param s: str
    :type s: ```str```
    """
    return s


class TestGenUtils(TestCase):
    """Test class for cdd.gen_utils"""

    def test_get_input_mapping_from_path(self) -> None:
        """test `get_input_mapping_from_path`"""
        self.assertEqual(f(""), "")
        name_to_node = get_input_mapping_from_path(
            "function", "cdd.tests.test_compound", "test_gen_utils"
        )
        self.assertEqual(len(name_to_node), 1)
        self.assertIn("f", name_to_node)
        self.assertIsInstance(name_to_node["f"], dict)


unittest_main()



"""
Shared utility functions for SQLalchemy
"""

import ast
from ast import Call, Expr, Load, Name, Subscript, Tuple, keyword
from operator import attrgetter

import cdd.compound.openapi.utils.emit_utils
from cdd.shared.ast_utils import get_value, set_value
from cdd.shared.pure_utils import PY_GTE_3_9, rpartial
from cdd.shared.source_transformer import to_code


def update_args_infer_typ_sqlalchemy(_param, args, name, nullable, x_typ_sql):
    """
    :param _param: Param with typ
    :type _param: ```dict```

    :param args:
    :type args: ```List```

    :param name:
    :type name: ```str```

    :param nullable:
    :type nullable: ```Optional[bool]```

    :param x_typ_sql:
    :type x_typ_sql: ```dict```

    :rtype: ```bool```
    """
    if _param["typ"].startswith("Optional["):
        _param["typ"] = _param["typ"][len("Optional[") : -1]
        nullable = True
    if "Literal[" in _param["typ"]:
        parsed_typ = get_value(ast.parse(_param["typ"]).body[0])
        assert (
            parsed_typ.value.id == "Literal"
        ), "Only basic Literal support is implemented, not {}".format(
            parsed_typ.value.id
        )
        args.append(
            Call(
                func=Name("Enum", Load()),
                args=get_value(parsed_typ.slice).elts,
                keywords=[
                    ast.keyword(arg="name", value=set_value(name), identifier=None)
                ],
                expr=None,
                expr_func=None,
            )
        )
    elif _param["typ"].startswith("List["):
        after_generic = _param["typ"][len("List[") :]
        if "struct" in after_generic:  # "," in after_generic or
            name = Name(id="JSON", ctx=Load())
        else:
            list_typ = ast.parse(_param["typ"]).body[0]
            assert isinstance(
                list_typ, Expr
            ), "Expected `Expr` got `{type_name}`".format(
                type_name=type(list_typ).__name__
            )
            assert isinstance(
                list_typ.value, Subscript
            ), "Expected `Subscript` got `{type_name}`".format(
                type_name=type(list_typ.value).__name__
            )
            name = next(
                filter(rpartial(isinstance, Name), ast.walk(list_typ.value.slice)), None
            )
            assert name is not None, "Could not find a type in {!r}".format(
                to_code(list_typ.value.slice)
            )
        args.append(
            Call(
                func=Name(id="ARRAY", ctx=Load()),
                args=[
                    Name(
                        id=cdd.compound.openapi.utils.emit_utils.typ2column_type.get(
                            name.id, name.id
                        ),
                        ctx=Load(),
                    )
                ],
                keywords=[],
                expr=None,
                expr_func=None,
            )
        )
    elif (
        "items" in _param
        and _param["items"].get("type", False)
        in cdd.compound.openapi.utils.emit_utils.typ2column_type
    ):
        args.append(
            Call(
                func=Name(id="ARRAY", ctx=Load()),
                args=[
                    Name(
                        id=cdd.compound.openapi.utils.emit_utils.typ2column_type[
                            _param["items"]["type"]
                        ],
                        ctx=Load(),
                    )
                ],
                keywords=[],
                expr=None,
                expr_func=None,
            )
        )
    elif _param.get("typ").startswith("Union["):
        # Hack to remove the union type. Enum parse seems to be incorrect?
        union_typ = ast.parse(_param["typ"]).body[0]
        assert isinstance(
            union_typ.value, Subscript
        ), "Expected `Subscript` got `{type_name}`".format(
            type_name=type(union_typ.value).__name__
        )
        union_typ_tuple = (
            union_typ.value.slice if PY_GTE_3_9 else union_typ.value.slice.value
        )
        assert isinstance(
            union_typ_tuple, Tuple
        ), "Expected `Tuple` got `{type_name}`".format(
            type_name=type(union_typ_tuple).__name__
        )
        assert (
            len(union_typ_tuple.elts) == 2
        ), "Expected length of 2 got `{tuple_len}`".format(
            tuple_len=len(union_typ_tuple.elts)
        )
        left, right = map(attrgetter("id"), union_typ_tuple.elts)
        args.append(
            Name(
                cdd.compound.openapi.utils.emit_utils.typ2column_type.get(right, right)
                if left in cdd.compound.openapi.utils.emit_utils.typ2column_type
                else cdd.compound.openapi.utils.emit_utils.typ2column_type.get(
                    left, left
                ),
                Load(),
            )
        )
    else:
        type_name = (
            x_typ_sql["type"]
            if "type" in x_typ_sql
            else cdd.compound.openapi.utils.emit_utils.typ2column_type.get(
                _param["typ"], _param["typ"]
            )
        )
        args.append(
            Call(
                func=Name(type_name, Load()),
                args=list(map(set_value, x_typ_sql.get("type_args", iter(())))),
                keywords=[
                    keyword(arg=arg, value=set_value(val))
                    for arg, val in x_typ_sql.get("type_kwargs", dict()).items()
                ],
                expr=None,
                expr_func=None,
            )
            if "type_args" in x_typ_sql or "type_kwargs" in x_typ_sql
            else Name(type_name, Load())
        )
    return nullable


__all__ = ["update_args_infer_typ_sqlalchemy"]



""" Tests for cst """

from os import extsep, path
from unittest import TestCase

from cdd.shared.cst import cst_parse
from cdd.tests.mocks.cst import cstify_cst
from cdd.tests.utils_for_tests import unittest_main


class TestCst(TestCase):
    """Test class for cst"""

    def test_cstify_file(self) -> None:
        """Tests that `handle_multiline_comment` can do its namesake"""
        with open(
            path.join(
                path.dirname(__file__),
                "../mocks",
                "cstify{extsep}py".format(extsep=extsep),
            ),
            "rt",
        ) as f:
            cst = cst_parse(f.read())

        self.assertTupleEqual(cst, cstify_cst)


unittest_main()



"""
Pydantic `class` parser

https://pydantic-docs.helpmanual.io/usage/schema/
"""

from functools import partial

import cdd.class_.parse

pydantic = partial(cdd.class_.parse.class_, infer_type=True)

__all__ = ["pydantic"]



"""
Tests for `cdd.emit.argparse`
"""

from copy import deepcopy
from unittest import TestCase

import cdd.argparse_function.emit
import cdd.argparse_function.parse
import cdd.class_.emit
import cdd.class_.parse
import cdd.docstring.emit
import cdd.function.emit
import cdd.function.parse
import cdd.json_schema.emit
import cdd.shared.emit.file
import cdd.sqlalchemy.emit
from cdd.tests.mocks.argparse import (
    argparse_func_action_append_ast,
    argparse_func_ast,
    argparse_func_torch_nn_l1loss_ast,
    argparse_func_with_body_ast,
    argparse_function_google_tf_tensorboard_ast,
)
from cdd.tests.mocks.classes import (
    class_ast,
    class_google_tf_tensorboard_ast,
    class_nargs_ast,
)
from cdd.tests.mocks.ir import class_torch_nn_l1loss_ir
from cdd.tests.utils_for_tests import reindent_docstring, run_ast_test, unittest_main


class TestEmitArgparse(TestCase):
    """Tests emission"""

    def test_to_argparse(self) -> None:
        """
        Tests whether `to_argparse` produces `argparse_func_ast` given `class_ast`
        """
        run_ast_test(
            self,
            gen_ast=cdd.argparse_function.emit.argparse_function(
                cdd.class_.parse.class_(class_ast),
                emit_default_doc=False,
            ),
            gold=argparse_func_ast,
        )

    def test_to_argparse_func_nargs(self) -> None:
        """
        Tests whether an argparse function is generated with `action="append"` set properly
        """
        run_ast_test(
            self,
            gen_ast=cdd.argparse_function.emit.argparse_function(
                cdd.class_.parse.class_(class_nargs_ast),
                emit_default_doc=False,
                function_name="set_cli_action_append",
            ),
            gold=argparse_func_action_append_ast,
        )

    def test_to_argparse_google_tf_tensorboard(self) -> None:
        """
        Tests whether `to_argparse` produces `argparse_function_google_tf_tensorboard_ast`
                                    given `class_google_tf_tensorboard_ast`
        """
        run_ast_test(
            self,
            gen_ast=cdd.argparse_function.emit.argparse_function(
                cdd.class_.parse.class_(
                    class_google_tf_tensorboard_ast, merge_inner_function="__init__"
                ),
                emit_default_doc=False,
                word_wrap=False,
            ),
            gold=argparse_function_google_tf_tensorboard_ast,
        )

    def test_from_argparse_with_extra_body_to_argparse_with_extra_body(self) -> None:
        """Tests if this can make the roundtrip from a full argparse function to a argparse full function"""

        ir = cdd.argparse_function.parse.argparse_ast(argparse_func_with_body_ast)
        func = cdd.argparse_function.emit.argparse_function(
            ir, emit_default_doc=False, word_wrap=True
        )
        run_ast_test(
            self, *map(reindent_docstring, (func, argparse_func_with_body_ast))
        )

    def test_from_torch_ir_to_argparse(self) -> None:
        """Tests if emission of class from torch IR is as expected"""

        func = cdd.argparse_function.emit.argparse_function(
            deepcopy(class_torch_nn_l1loss_ir),
            emit_default_doc=False,
            wrap_description=False,
            word_wrap=False,
        )
        run_ast_test(
            self,
            func,
            argparse_func_torch_nn_l1loss_ast,
        )


unittest_main()



""" Tests for doctrans_utils """

from ast import Expr, Load, Module, Name, fix_missing_locations, get_docstring
from collections import deque
from copy import deepcopy
from os import path
from os.path import extsep
from unittest import TestCase
from unittest.mock import MagicMock, patch

from cdd.compound.doctrans_utils import (
    DocTrans,
    clear_annotation,
    doctransify_cst,
    has_type_annotations,
)
from cdd.shared.ast_utils import annotate_ancestry, set_value
from cdd.shared.pure_utils import omit_whitespace
from cdd.shared.source_transformer import ast_parse
from cdd.tests.mocks.cst import cstify_cst
from cdd.tests.mocks.doctrans import (
    ann_assign_with_annotation,
    assign_with_type_comment,
    class_with_internal_annotated,
    class_with_internal_type_commented_and_docstring_typed,
    function_type_annotated,
    function_type_in_docstring,
)
from cdd.tests.mocks.methods import function_google_tf_ops_losses__safe_mean_ast
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


class TestDocTransUtils(TestCase):
    """Test class for doctrans_utils.py"""

    def test_has_type_annotations(self) -> None:
        """Tests has_type_annotations"""

        self.assertTrue(has_type_annotations(ast_parse("a: int = 5")))
        self.assertFalse(has_type_annotations(ast_parse("a = 5")))
        self.assertTrue(has_type_annotations(ast_parse("def a() -> None: pass")))
        self.assertFalse(has_type_annotations(ast_parse("def a(): pass")))

    def test_doctrans_function_from_annotated_to_docstring(self) -> None:
        """Tests `DocTrans` converts type annotated function to docstring function"""

        original_node = annotate_ancestry(deepcopy(function_type_annotated))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=False,
            existing_type_annotations=True,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)

        run_ast_test(self, gen_ast, gold=function_type_in_docstring)

    def test_doctrans_function_from_docstring_to_annotated(self) -> None:
        """Tests `DocTrans` converts docstring function to type annotated function"""
        original_node = annotate_ancestry(deepcopy(function_type_in_docstring))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)

        run_ast_test(self, gen_ast, gold=function_type_annotated)

    def test_doctrans_assign_to_annassign(self) -> None:
        """
        Tests that `Assign` converts to `AnnAssign`
        """
        original_node = annotate_ancestry(deepcopy(assign_with_type_comment))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        run_ast_test(
            self,
            gen_ast=doc_trans.visit(original_node),
            gold=ann_assign_with_annotation,
        )

    def test_doctrans_annassign_to_assign(self) -> None:
        """
        Tests that `AnnAssign` converts to `Assign`
        """
        original_node = annotate_ancestry(deepcopy(ann_assign_with_annotation))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=False,
            existing_type_annotations=True,
            whole_ast=original_node,
        )
        run_ast_test(
            self, gen_ast=doc_trans.visit(original_node), gold=assign_with_type_comment
        )

    def test_doctrans_annassign_to_assign_with_clearing_type_annotations(self) -> None:
        """
        Tests that `AnnAssign` converts to `Assign`
        """
        original_node = annotate_ancestry(deepcopy(ann_assign_with_annotation))
        original_node.type_comment = "NEVER SEE THIS"
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)
        run_ast_test(self, gen_ast=gen_ast, gold=ann_assign_with_annotation)

    def test_doctrans_assign_to_assign(self) -> None:
        """
        Tests that `AnnAssign` converts to `Assign`
        """
        original_node = annotate_ancestry(deepcopy(assign_with_type_comment))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=False,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)
        run_ast_test(self, gen_ast=gen_ast, gold=assign_with_type_comment)

    def test_class_with_internal_annotated(self) -> None:
        """Tests that class, function, and class variable hierarchy is correctly annotated handles the ident case"""
        original_node = annotate_ancestry(deepcopy(class_with_internal_annotated))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)
        run_ast_test(self, gen_ast=gen_ast, gold=class_with_internal_annotated)

    def test__get_ass_typ(self) -> None:
        """Tests that _get_ass_typ returns when location isn't set"""
        original_node = annotate_ancestry(deepcopy(assign_with_type_comment))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        del original_node._location
        self.assertEqual(
            doc_trans._get_ass_typ(original_node),
            Name("int", Load()).id,
        )

    def test_class_with_internal_converts_to_annotated(self) -> None:
        """Tests that class, function, and class variable hierarchy is correctly converts to annotated"""
        original_node = annotate_ancestry(
            deepcopy(class_with_internal_type_commented_and_docstring_typed)
        )
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)
        run_ast_test(self, gen_ast=gen_ast, gold=class_with_internal_annotated)

    def test_class_annotated_converts_to_type_commented_and_docstring_typed(
        self,
    ) -> None:
        """Tests that class, function, and class variable hierarchy is correctly converted to annotated"""
        original_node = annotate_ancestry(deepcopy(class_with_internal_annotated))
        doc_trans = DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=False,
            existing_type_annotations=True,
            whole_ast=original_node,
        )
        gen_ast = doc_trans.visit(original_node)
        run_ast_test(
            self,
            gen_ast=gen_ast,
            gold=class_with_internal_type_commented_and_docstring_typed,
        )

    def test_module_docstring(self) -> None:
        """Tests that module gets the right new docstring"""
        module_node = Module(
            body=[Expr(set_value("\nModule\n"))], stmt=None, type_ignores=[]
        )
        original = deepcopy(module_node)
        DocTrans(
            docstring_format="rest",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=module_node,
        )
        run_ast_test(self, gen_ast=module_node, gold=original)

    def test_empty_types(self) -> None:
        """Tests function_google_tf_ops_losses__safe_mean_ast (which has empty arg types)"""
        original = Module(
            body=[
                fix_missing_locations(
                    deepcopy(function_google_tf_ops_losses__safe_mean_ast)
                )
            ],
            stmt=None,
            type_ignores=[],
        )
        doc_trans = DocTrans(
            docstring_format="google",
            word_wrap=True,
            type_annotations=True,
            existing_type_annotations=False,
            whole_ast=deepcopy(original),
        )

        # Reindent docstrings
        for body in original.body[0], doc_trans.whole_ast.body[0]:
            body.body[0] = Expr(
                set_value(omit_whitespace(get_docstring(body, clean=True)))
            )

        run_ast_test(self, gen_ast=doc_trans.whole_ast, gold=original)

    def test_clear_annotation(self) -> None:
        """Tests that `clear_annotation` clears correctly"""
        node_cls = type("Node", tuple(), {"annotation": None, "type_comment": None})
        node = node_cls()
        node.annotation = 5
        node.type_comment = 6
        clear_annotation(node)
        deque(
            (
                self.assertIsNone(getattr(node, attr))
                for attr in dir(node_cls)
                if not attr.startswith("_")
            ),
            maxlen=0,
        )

    def test_doctransify_cst(self) -> None:
        """
        Tests that `doctransify_cst` calls the node potentially-augmenting functions right number of times
        """
        cst_list = list(deepcopy(cstify_cst))
        with open(
            path.join(
                path.dirname(path.dirname(__file__)),
                "mocks",
                "cstify{extsep}py".format(extsep=extsep),
            ),
            "rt",
        ) as f:
            ast_mod = ast_parse(f.read(), skip_docstring_remit=True)

        fake_maybe_replace_doc_str_in_function_or_class = MagicMock()
        with patch(
            "cdd.shared.ast_cst_utils.maybe_replace_doc_str_in_function_or_class",
            fake_maybe_replace_doc_str_in_function_or_class,
        ):
            doctransify_cst(cst_list, ast_mod)
        self.assertTrue(fake_maybe_replace_doc_str_in_function_or_class.called)
        self.assertEqual(fake_maybe_replace_doc_str_in_function_or_class.call_count, 6)


unittest_main()



"""
pkg_utils
"""

from distutils.sysconfig import get_python_lib


def relative_filename(filename, remove_hints=tuple()):
    """
    Remove all the paths which are not relevant

    :param filename: Filename
    :type filename: ```str```

    :param remove_hints: Hints as to what can be removed
    :type remove_hints: ```Tuple[str]```

    :return: Relative path (if derived) else original
    :rtype: ```str```
    """
    _filename = filename.casefold()
    for elem in remove_hints + (get_python_lib(), get_python_lib(prefix="")):
        if _filename.startswith(elem.casefold()):
            return filename[len(elem) + 1 :]
    return filename


__all__ = ["relative_filename"]



"""
Function/method emitter
"""

import ast
from ast import Expr, FunctionDef, Load, Name, Return, arguments

from cdd.docstring.emit import docstring
from cdd.shared.ast_utils import maybe_type_comment, set_arg, set_value
from cdd.shared.emit.utils.emitter_utils import get_internal_body
from cdd.shared.pure_utils import PY3_8, none_types, simple_types


def function(
    intermediate_repr,
    function_name,
    function_type,
    word_wrap=True,
    emit_default_doc=False,
    docstring_format="rest",
    indent_level=2,
    emit_separating_tab=PY3_8,
    type_annotations=True,
    emit_as_kwonlyargs=True,
    emit_original_whitespace=False,
):
    """
    Construct a function from our IR

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param function_name: name of function_def
    :type function_name: ```Optional[str]```

    :param function_type: Type of function, static is static or global method, others just become first arg
    :type function_type: ```Optional[Literal['self', 'cls', 'static']]```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :param indent_level: docstring indentation level whence: 0=no_tabs, 1=one tab; 2=two tabs
    :type indent_level: ```int```

    :param emit_separating_tab: docstring decider for whether to put a tab between :param and return and desc
    :type emit_separating_tab: ```bool```

    :param type_annotations: True to have type annotations (3.6+), False to place in docstring
    :type type_annotations: ```bool```

    :param emit_as_kwonlyargs: Whether argument(s) emitted must be keyword only
    :type emit_as_kwonlyargs: ```bool```

    :param emit_original_whitespace: Whether to emit an original whitespace (in docstring) or strip it out
    :type emit_original_whitespace: ```bool```

    :return: AST node for function definition
    :rtype: ```FunctionDef```
    """
    params_no_kwargs = tuple(
        filter(
            lambda param: not param[0].endswith("kwargs"),
            intermediate_repr["params"].items(),
        )
    )

    function_name = function_name or intermediate_repr["name"]
    function_type = function_type or intermediate_repr["type"]

    args = (
        [] if function_type in frozenset((None, "static")) else [set_arg(function_type)]
    )
    from cdd.shared.emit.utils.emitter_utils import ast_parse_fix

    args_from_params = list(
        map(
            lambda param: set_arg(
                annotation=(
                    Name(param[1]["typ"], Load())
                    if param[1]["typ"] in simple_types
                    else ast_parse_fix(param[1]["typ"])
                )
                if type_annotations and "typ" in param[1]
                else None,
                arg=param[0],
            ),
            params_no_kwargs,
        ),
    )
    defaults_from_params = list(
        map(
            lambda param: set_value(None)
            if param[1].get("default") in none_types
            else set_value(param[1].get("default")),
            params_no_kwargs,
        )
    )
    if emit_as_kwonlyargs:
        kwonlyargs, kw_defaults, defaults = args_from_params, defaults_from_params, []
    else:
        kwonlyargs, kw_defaults, defaults = [], [], defaults_from_params
        args += args_from_params

    internal_body = get_internal_body(
        target_name=function_name,
        target_type=function_type,
        intermediate_repr=intermediate_repr,
    )
    return_val = (
        Return(
            value=ast.parse(
                intermediate_repr["returns"]["return_type"]["default"].strip("`")
            )
            .body[0]
            .value,
            expr=None,
        )
        if (intermediate_repr.get("returns") or {"return_type": {}})["return_type"].get(
            "default"
        )
        else None
    )

    return FunctionDef(
        args=arguments(
            args=args,
            defaults=defaults,
            kw_defaults=kw_defaults,
            kwarg=next(
                map(
                    lambda param: set_arg(param[0]),
                    filter(
                        lambda param: param[0].endswith("kwargs"),
                        intermediate_repr["params"].items(),
                    ),
                ),
                None,
            ),
            kwonlyargs=kwonlyargs,
            posonlyargs=[],
            vararg=None,
            arg=None,
        ),
        body=list(
            filter(
                None,
                (
                    Expr(
                        set_value(
                            docstring(
                                intermediate_repr,
                                docstring_format=docstring_format,
                                emit_default_doc=emit_default_doc,
                                emit_original_whitespace=emit_original_whitespace,
                                emit_separating_tab=emit_separating_tab,
                                emit_types=not type_annotations,
                                indent_level=indent_level,
                                word_wrap=word_wrap,
                            )
                        )
                    ),
                    *(
                        internal_body[:-1]
                        if internal_body
                        and isinstance(internal_body[-1], Return)
                        and return_val
                        else internal_body
                    ),
                    return_val,
                ),
            )
        ),
        decorator_list=[],
        name=function_name,
        returns=(
            ast.parse(intermediate_repr["returns"]["return_type"]["typ"]).body[0].value
            if type_annotations
            and (intermediate_repr.get("returns") or {"return_type": {}})[
                "return_type"
            ].get("typ")
            else None
        ),
        lineno=None,
        arguments_args=None,
        identifier_name=None,
        stmt=None,
        **maybe_type_comment
    )


__all__ = ["function"]



"""
`class` emitter
"""

import ast
from ast import ClassDef, Expr, FunctionDef, Load, Name
from collections import OrderedDict
from functools import partial
from itertools import chain

from cdd.class_.utils.emit_utils import RewriteName
from cdd.docstring.emit import docstring
from cdd.function.utils.emit_utils import _make_call_meth
from cdd.shared.ast_utils import param2ast, set_value
from cdd.shared.pure_utils import rpartial


def class_(
    intermediate_repr,
    emit_call=False,
    class_name=None,
    class_bases=("object",),
    decorator_list=None,
    word_wrap=True,
    docstring_format="rest",
    emit_original_whitespace=False,
    emit_default_doc=False,
):
    """
    Construct a class

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param emit_call: Whether to emit a `__call__` method from the `_internal` IR subdict
    :type emit_call: ```bool```

    :param class_name: name of class
    :type class_name: ```str```

    :param class_bases: bases of class (the generated class will inherit these)
    :type class_bases: ```Iterable[str]```

    :param decorator_list: List of decorators
    :type decorator_list: ```Optional[Union[List[Str], List[]]]```

    :param word_wrap: Whether to word-wrap. Set `DOCTRANS_LINE_LENGTH` to configure length.
    :type word_wrap: ```bool```

    :param docstring_format: Format of docstring
    :type docstring_format: ```Literal['rest', 'numpydoc', 'google']```

    :param emit_original_whitespace: Whether to emit original whitespace or strip it out (in docstring)
    :type emit_original_whitespace: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Class AST
    :rtype: ```ClassDef```
    """
    assert isinstance(
        intermediate_repr, dict
    ), "Expected `dict` got `{type_name}`".format(
        type_name=type(intermediate_repr).__name__
    )
    assert class_name or intermediate_repr["name"], "Class has no name"

    returns = (
        intermediate_repr["returns"]
        if "return_type" in ((intermediate_repr or {}).get("returns") or iter(()))
        else OrderedDict()
    )
    if returns:
        intermediate_repr["params"].update(returns)
        del intermediate_repr["returns"]

    internal_body = intermediate_repr.get("_internal", {}).get("body", [])
    # TODO: Add correct classmethod/staticmethod to decorate function using `annotate_ancestry` and first-field checks
    # Such that the `self.` or `cls.` rewrite only applies to non-staticmethods
    # assert internal_body, "Expected `internal_body` to have contents"
    param_names = (
        frozenset(intermediate_repr["params"].keys())
        if "params" in intermediate_repr
        else None
    )
    if param_names:
        if internal_body:
            internal_body = list(
                map(
                    ast.fix_missing_locations,
                    map(RewriteName(param_names).visit, internal_body),
                )
            )
        elif (returns or {"return_type": None}).get("return_type") is not None:
            internal_body = returns["return_type"]

    indent_level = 1

    _emit_docstring = partial(
        docstring,
        docstring_format=docstring_format,
        indent_level=indent_level,
        emit_default_doc=emit_default_doc,
        emit_separating_tab=True,
        emit_types=False,
        word_wrap=word_wrap,
    )
    return ClassDef(
        bases=list(map(rpartial(Name, Load()), class_bases)),
        body=list(
            filter(
                None,
                chain.from_iterable(
                    (
                        (
                            (lambda ds: None if ds is None else Expr(set_value(ds)))(
                                _emit_docstring(
                                    {
                                        k: intermediate_repr[k]
                                        for k in intermediate_repr
                                        if k != "_internal"
                                    },
                                    emit_original_whitespace=emit_original_whitespace,
                                    purpose="class",
                                ).rstrip()
                                or None
                            ),
                        ),
                        map(
                            param2ast,
                            (intermediate_repr.get("params") or OrderedDict()).items(),
                        ),
                        iter(
                            (
                                (
                                    internal_body[0]
                                    if len(internal_body) == 1
                                    and isinstance(internal_body[0], FunctionDef)
                                    and internal_body[0].name == "__call__"
                                    else _make_call_meth(
                                        internal_body,
                                        returns["return_type"]["default"]
                                        if "default"
                                        in (
                                            (returns or {"return_type": iter(())}).get(
                                                "return_type"
                                            )
                                            or iter(())
                                        )
                                        else None,
                                        param_names,
                                        docstring_format=docstring_format,
                                        word_wrap=word_wrap,
                                    ),
                                )
                                or iter(())
                            )
                            if emit_call and internal_body
                            else iter(())
                        ),
                    ),
                ),
            )
        )
        or [Expr(set_value(""))],  # empty body will cause syntax error
        decorator_list=list(map(rpartial(Name, Load()), decorator_list))
        if decorator_list
        else [],
        keywords=[],
        name=class_name or intermediate_repr["name"],
        expr=None,
        identifier_name=None,
    )


__all__ = ["class_"]



"""
Tests for the Intermediate Representation produced by the parsers
"""

from collections import OrderedDict
from unittest import TestCase

import cdd.docstring.parse
from cdd.tests.mocks.docstrings import (
    docstring_keras_rmsprop_class_str,
    docstring_keras_rmsprop_method_str,
    docstring_reduction_v2_str,
)
from cdd.tests.mocks.ir import (
    docstring_keras_rmsprop_class_ir,
    docstring_keras_rmsprop_method_ir,
)
from cdd.tests.utils_for_tests import unittest_main


class TestParseDocstring(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_docstring_docstring_reduction_v2_str(self) -> None:
        """
        Test that the nonmatching docstring doesn't fill out params
        """
        ir = cdd.docstring.parse.docstring(docstring_reduction_v2_str)
        self.assertEqual(ir["params"], OrderedDict())
        self.assertEqual(ir["returns"], None)

    def test_from_docstring_keras_rmsprop_class_str(self) -> None:
        """Tests IR from docstring_keras_rmsprop_class_str"""

        self.assertDictEqual(
            cdd.docstring.parse.docstring(docstring_keras_rmsprop_class_str),
            docstring_keras_rmsprop_class_ir,
        )

    def test_from_docstring_keras_rmsprop_class_method_str(self) -> None:
        """Tests IR from docstring_keras_rmsprop_method_str"""

        self.assertDictEqual(
            cdd.docstring.parse.docstring(docstring_keras_rmsprop_method_str),
            docstring_keras_rmsprop_method_ir,
        )


unittest_main()



""" Tests for exmod subcommand """

from ast import ClassDef
from functools import partial
from io import StringIO
from itertools import chain, groupby
from operator import add, itemgetter
from os import environ, listdir, mkdir, path, walk
from os.path import extsep
from subprocess import DEVNULL, call
from sys import executable, platform
from tempfile import TemporaryDirectory
from unittest import TestCase
from unittest.mock import patch

import cdd.class_.parse
from cdd.compound.exmod import exmod
from cdd.shared.pkg_utils import relative_filename
from cdd.shared.pure_utils import ENCODING, INIT_FILENAME, rpartial, unquote
from cdd.shared.source_transformer import ast_parse
from cdd.tests.mocks import imports_header
from cdd.tests.mocks.classes import class_str
from cdd.tests.mocks.exmod import setup_py_mock
from cdd.tests.utils_for_tests import unittest_main


class TestExMod(TestCase):
    """Test class for exmod.py"""

    def test_exmod(self) -> None:
        """Tests `exmod`"""

        try:
            with TemporaryDirectory(
                prefix="gold", suffix="gold"
            ) as existent_module_dir, TemporaryDirectory(
                prefix="gen", suffix="gen"
            ) as new_module_dir:
                self._create_fs(existent_module_dir)
                # sys.path.insert(0, existent_module_dir)
                self._pip(["install", "."], existent_module_dir)

                exmod(
                    module=self.module_name,
                    emit_name="class",
                    blacklist=tuple(),
                    whitelist=tuple(),
                    mock_imports=True,
                    output_directory=new_module_dir,
                    no_word_wrap=None,
                    dry_run=False,
                )
                self._check_emission(new_module_dir)
        finally:
            # sys.path.remove(existent_module_dir)
            self._pip(["uninstall", "-y", self.module_name])

    def test_exmod_blacklist(self) -> None:
        """Tests `exmod` blacklist"""

        try:
            with TemporaryDirectory(
                prefix="gold", suffix="gold"
            ) as existent_module_dir, TemporaryDirectory(
                prefix="gen", suffix="gen"
            ) as new_module_dir:
                self._create_fs(existent_module_dir)
                self._pip(["install", "."], existent_module_dir)
                exmod(
                    module=self.module_name,
                    emit_name="class",
                    blacklist=(".".join((self.module_name,) * 2),),
                    whitelist=tuple(),
                    mock_imports=True,
                    output_directory=new_module_dir,
                    no_word_wrap=None,
                    dry_run=False,
                )
                self.assertListEqual(listdir(new_module_dir), [])
        finally:
            self._pip(["uninstall", "-y", self.module_name])

    def test_exmod_whitelist(self) -> None:
        """Tests `exmod` whitelist"""

        try:
            with TemporaryDirectory(
                prefix="gold", suffix="gold"
            ) as existent_module_dir, TemporaryDirectory(
                prefix="gen", suffix="gen"
            ) as new_module_dir:
                self._create_fs(existent_module_dir)
                self._pip(["install", "."], existent_module_dir)
                existent_module_name = path.basename(existent_module_dir)
                exmod(
                    module=self.module_name,
                    emit_name="class",
                    blacklist=tuple(),
                    whitelist=(".".join((existent_module_name,) * 2),),
                    mock_imports=True,
                    output_directory=new_module_dir,
                    no_word_wrap=None,
                    dry_run=False,
                )

                gen, gold = map(
                    sorted,
                    (
                        map(
                            lambda p: (
                                lambda _p: path.join("gold", _p.partition(path.sep)[2])
                                if _p.startswith("gold") and _p.count("gold") == 2
                                else _p
                            )(p.partition(path.sep)[2]),
                            (
                                path.join(dirpath, filename)[
                                    len(new_module_dir + path.sep) :
                                ]
                                for (dirpath, dirnames, filenames) in walk(
                                    new_module_dir
                                )
                                for filename in filenames
                            ),
                        ),
                        (
                            INIT_FILENAME,
                            *chain.from_iterable(
                                map(
                                    lambda i: map(
                                        partial(
                                            path.join,
                                            self.module_hierarchy[i][1],
                                        ),
                                        (
                                            "{name}{sep}py".format(
                                                name=self.module_hierarchy[i][0],
                                                sep=extsep,
                                            ),
                                            INIT_FILENAME,
                                        ),
                                    ),
                                    range(len(self.module_hierarchy)),
                                ),
                            ),
                        ),
                    ),
                )
                all_tests_running = len(gold) == 7
                if all_tests_running:
                    gold = list(
                        chain.from_iterable(
                            (
                                gold[:1],
                                chain.from_iterable(
                                    map(
                                        lambda p: (p, p),
                                        map(
                                            partial(path.join, "gold", "parent_dir"),
                                            (
                                                INIT_FILENAME,
                                                *map(
                                                    partial(path.join, "child_dir"),
                                                    (
                                                        INIT_FILENAME,
                                                        "child{sep}py".format(
                                                            sep=extsep
                                                        ),
                                                        path.join(
                                                            "grandchild_dir",
                                                            INIT_FILENAME,
                                                        ),
                                                        path.join(
                                                            "grandchild_dir",
                                                            "grandchild{sep}py".format(
                                                                sep=extsep
                                                            ),
                                                        ),
                                                    ),
                                                ),
                                                "parent{sep}py".format(sep=extsep),
                                            ),
                                        ),
                                    )
                                ),
                                gold[1:],
                            )
                        )
                    )

                self.assertListEqual(gen, gold)
        finally:
            self._pip(["uninstall", "-y", self.module_name])

    def test_exmod_module_directory(self) -> None:
        """Tests `exmod` module whence directory"""

        with TemporaryDirectory() as tempdir, self.assertRaises(ModuleNotFoundError):
            exmod(
                module=tempdir,
                emit_name="cool_name",
                blacklist=tuple(),
                whitelist=tuple(),
                mock_imports=True,
                output_directory=path.join(tempdir, "nonexistent"),
                no_word_wrap=None,
                dry_run=False,
            )

    def test_exmod_no_module(self) -> None:
        """Tests that ModuleNotFound error is raised when module is not installed"""
        with TemporaryDirectory() as tempdir, self.assertRaises(ModuleNotFoundError):
            exmod(
                module="fubar",
                emit_name="uncool_name",
                blacklist=tuple(),
                whitelist=tuple(),
                mock_imports=True,
                output_directory=path.join(tempdir, "nonexistent"),
                no_word_wrap=None,
                dry_run=False,
            )

    def test_exmod_output_directory_nonexistent(self) -> None:
        """Tests `exmod` module whence directory does not exist"""

        with TemporaryDirectory() as tempdir, self.assertRaises(ModuleNotFoundError):
            output_directory = path.join(tempdir, "stuff")
            self.assertFalse(path.isdir(output_directory))
            exmod(
                module=output_directory,
                emit_name=None,
                blacklist=tuple(),
                whitelist=tuple(),
                mock_imports=True,
                output_directory=output_directory,
                no_word_wrap=None,
                dry_run=False,
            )

    def test_exmod_dry_run(self) -> None:
        """Tests `exmod` dry_run"""

        try:
            with TemporaryDirectory(
                prefix="gold", suffix="gold"
            ) as existent_module_dir, TemporaryDirectory(
                prefix="gen", suffix="gen"
            ) as new_module_dir:
                self._create_fs(existent_module_dir)
                self._pip(["install", "."], existent_module_dir)

                with patch("sys.stdout", new_callable=StringIO) as f:
                    exmod(
                        module=self.module_name,
                        emit_name="class",
                        blacklist=tuple(),
                        whitelist=tuple(),
                        mock_imports=True,
                        output_directory=new_module_dir,
                        no_word_wrap=None,
                        dry_run=True,
                    )
                    r = f.getvalue()

                result = dict(
                    map(
                        lambda k_v: (
                            k_v[0],
                            tuple(
                                sorted(
                                    set(
                                        map(
                                            partial(
                                                relative_filename,
                                                remove_hints=(
                                                    (
                                                        lambda directory: "{directory}{sep}".format(
                                                            directory=unquote(
                                                                repr(directory)
                                                            ),
                                                            sep=path.sep,
                                                        )
                                                        if platform == "win32"
                                                        else directory
                                                    )(
                                                        path.join(
                                                            new_module_dir,
                                                            path.basename(
                                                                new_module_dir
                                                            ),
                                                        ),
                                                    ),
                                                ),
                                            ),
                                            map(unquote, map(itemgetter(1), k_v[1])),
                                        )
                                    )
                                )
                            ),
                        ),
                        groupby(
                            map(rpartial(str.split, "\t", 2), sorted(r.splitlines())),
                            key=itemgetter(0),
                        ),
                    )
                )

                all_tests_running = len(result["write"]) == 1

                key_counts = (
                    (("mkdir", 10), ("touch", 4), ("write", 1))
                    if all_tests_running
                    else (("mkdir", 7), ("touch", 4), ("write", 4))
                )

                for key, count in key_counts:
                    self.assertEqual(count, len(result[key]), key)

                gold_module_name = next(
                    map(
                        lambda p: p.partition(path.sep)[0],
                        filter(rpartial(str.startswith, "gold"), result["mkdir"]),
                    ),
                    path.basename(self.gold_dir),
                )
                gold = {
                    k: tuple(map(unquote, map(repr, v)))
                    for k, v in {
                        "mkdir": chain.from_iterable(
                            (
                                (new_module_dir,),
                                map(
                                    partial(path.join, gold_module_name),
                                    (
                                        self.module_hierarchy[0][1],
                                        self.module_hierarchy[1][1],
                                        self.module_hierarchy[2][1],
                                    ),
                                )
                                if all_tests_running
                                else iter(()),
                                (
                                    self.module_hierarchy[0][1],
                                    self.module_hierarchy[1][1],
                                    path.join(
                                        self.module_hierarchy[1][1],
                                        self.module_hierarchy[1][0],
                                    ),
                                    self.module_hierarchy[2][1],
                                    path.join(
                                        self.module_hierarchy[2][1],
                                        self.module_hierarchy[2][0],
                                    ),
                                    path.join(
                                        self.module_hierarchy[0][1],
                                        self.module_hierarchy[0][0],
                                    ),
                                ),
                            )
                        ),
                        "touch": (
                            INIT_FILENAME,
                            *map(
                                rpartial(path.join, INIT_FILENAME),
                                (
                                    self.module_hierarchy[0][1],
                                    self.module_hierarchy[1][1],
                                    self.module_hierarchy[2][1],
                                ),
                            ),
                        ),
                        "write": (
                            lambda write_block: tuple(write_block[:1])
                            if all_tests_running
                            else write_block
                        )(
                            (
                                INIT_FILENAME,
                                path.join(
                                    self.module_hierarchy[1][1],
                                    "{name}{extsep}py".format(
                                        name=self.module_hierarchy[1][0],
                                        extsep=extsep,
                                    ),
                                ),
                                path.join(
                                    self.module_hierarchy[2][1],
                                    "{name}{extsep}py".format(
                                        name=self.module_hierarchy[2][0],
                                        extsep=extsep,
                                    ),
                                ),
                                path.join(
                                    self.module_hierarchy[0][1],
                                    "{name}{extsep}py".format(
                                        name=self.module_hierarchy[0][0],
                                        extsep=extsep,
                                    ),
                                ),
                            )
                        ),
                    }.items()
                }

                self.assertDictEqual(result, gold)

                self._check_emission(new_module_dir, dry_run=True)
        finally:
            self._pip(["uninstall", "-y", self.module_name])

    def _create_fs(self, tempdir):
        """
        Populate filesystem from `tempdir` root with module hierarchy &etc. for later exposure (exmod)

        :param tempdir: Temporary directory
        :type tempdir: ```str```

        :return: tempdir
        :rtype: ```str```
        """
        self.module_name, self.gold_dir = path.basename(tempdir), tempdir
        self.parent_name, self.parent_dir = "parent", "parent_dir"
        self.child_name, self.child_dir = "child", path.join(
            self.parent_dir, "child_dir"
        )
        self.grandchild_name, self.grandchild_dir = "grandchild", path.join(
            self.child_dir, "grandchild_dir"
        )
        self.module_hierarchy = (
            (self.parent_name, self.parent_dir),
            (self.child_name, self.child_dir),
            (self.grandchild_name, self.grandchild_dir),
        )

        with open(
            path.join(tempdir, "setup{extsep}py".format(extsep=extsep)), "wt"
        ) as f:
            f.write(
                setup_py_mock.format(encoding=ENCODING, package_name=self.module_name)
            )

        open(path.join(tempdir, "README{extsep}md".format(extsep=extsep)), "a").close()
        mkdir(path.join(tempdir, self.module_name))
        with open(
            path.join(tempdir, self.module_name, INIT_FILENAME),
            "wt",
        ) as f:
            f.write(
                "{encoding}\n\n"
                "{imports}\n"
                "__author__ = {author!r}\n"
                "__version__ = {version!r}\n\n"
                "{all__}\n".format(
                    encoding=ENCODING,
                    imports="\n".join(
                        (
                            "import {module_name}.{other_imports}\n".format(
                                module_name=self.module_name,
                                other_imports="\nimport {module_name}.".format(
                                    module_name=self.module_name
                                ).join(
                                    map(
                                        rpartial(str.replace, path.sep, "."),
                                        map(itemgetter(1), self.module_hierarchy),
                                    )
                                ),
                            ),
                        )
                    ),
                    # module_name=self.module_name,
                    # parent_name=self.parent_name,
                    # cls_name="{name}Class".format(name=self.parent_name.title()),
                    author=environ.get("CDD_AUTHOR", "Samuel Marks"),
                    version=environ.get("CDD_VERSION", "0.0.0"),
                    all__="__all__ = {__all__!r}".format(
                        __all__=list(
                            map(
                                rpartial(add, "_dir"),
                                (
                                    self.parent_name,
                                    self.child_name,
                                    self.grandchild_name,
                                ),
                            )
                        )
                    ),
                )
            )

        for name, _folder in self.module_hierarchy:
            folder = path.join(tempdir, self.module_name, _folder)
            mkdir(folder)
            cls_name = "{name}Class".format(name=name.title())
            with open(
                path.join(folder, "__init__{extsep}py".format(extsep=extsep)), "wt"
            ) as f:
                f.write(
                    "{encoding}\n\n"
                    "from .{name} import {cls_name}\n\n"
                    "__all__ = [{cls_name!r}]\n".format(
                        encoding=ENCODING,
                        name=name,
                        cls_name=cls_name,
                    )
                )
            with open(
                path.join(folder, "{name}{extsep}py".format(name=name, extsep=extsep)),
                "wt",
            ) as f:
                f.write(
                    "{encoding}\n\n"
                    "{imports_header}\n"
                    "{class_str}\n\n"
                    "__all__ = [{cls_name!r}]\n".format(
                        encoding=ENCODING,
                        imports_header=imports_header,
                        class_str=class_str.replace("ConfigClass", cls_name),
                        cls_name=cls_name,
                    )
                )

        return tempdir

    def _check_emission(self, tempdir, dry_run=False):
        """
        Confirm whether emission conforms to gen by verifying their IRs are equivalent

        :param tempdir: Temporary directory
        :type tempdir: ```str```

        :param dry_run: Show what would be created; don't actually write to the filesystem
        :type dry_run: ```bool```
        """
        new_module_name = path.basename(tempdir)

        for name, folder in self.module_hierarchy:
            gen_folder = path.join(tempdir, new_module_name, folder)
            gold_folder = path.join(self.gold_dir, self.module_name, folder)

            def _open(_folder):
                """
                :param _folder: Folder to join on
                :type _folder: ```str``

                :return: Open IO
                :rtype: ```open```
                """
                return open(
                    path.join(
                        _folder, "{name}{extsep}py".format(name=name, extsep=extsep)
                    ),
                    "rt",
                )

            self.assertTrue(path.isdir(gold_folder))

            gen_is_dir = path.isdir(gen_folder)
            if dry_run:
                self.assertFalse(gen_is_dir)
            else:
                self.assertTrue(gen_is_dir)

                with _open(gen_folder) as gen, _open(gold_folder) as gold:
                    gen_ir, gold_ir = map(
                        lambda ir: ir["_internal"].__delitem__("original_doc_str")
                        or ir,
                        map(
                            lambda node: cdd.class_.parse.class_(
                                next(
                                    filter(
                                        rpartial(isinstance, ClassDef),
                                        (
                                            lambda node_read: ast_parse(  # print("%%%%%\n", node_read, ';') or
                                                node_read
                                            ).body
                                        )(node.read()),
                                    )
                                )
                            ),
                            (gen, gold),
                        ),
                    )
                    self.assertDictEqual(gold_ir, gen_ir)

    def _pip(self, pip_args, cwd=None):
        """
        Run `pip` with given args (and assert success).
        [Not using InstallCommand from pip anymore as its been deprecated (and now removed).]

        :param pip_args: Arguments to give pip
        :type pip_args: ```List[str]```

        :param cwd: Current working directory to run the command from. Defaults to current dir.
        :type cwd: ```Optional[str]```
        """
        self.assertEqual(
            call(
                [executable, "-m", "pip"] + pip_args,
                cwd=cwd,
                stdout=DEVNULL,
                stderr=DEVNULL,
            ),
            0,
            "EXIT_SUCCESS not reached",
        )


unittest_main()



"""
Tests for the Intermediate Representation produced by the SQLalchemy parsers
"""

import ast
from copy import deepcopy
from unittest import TestCase

import cdd.sqlalchemy.parse
from cdd.tests.mocks.ir import (
    intermediate_repr_no_default_sql_with_sql_types,
    intermediate_repr_node_pk,
)
from cdd.tests.mocks.sqlalchemy import (
    config_decl_base_ast,
    config_decl_base_str,
    config_tbl_with_comments_ast,
    config_tbl_with_comments_str,
    foreign_sqlalchemy_tbls_mod,
    foreign_sqlalchemy_tbls_str,
    node_pk_tbl_ass,
)
from cdd.tests.utils_for_tests import run_ast_test, unittest_main


class TestParseSqlAlchemy(TestCase):
    """
    Tests whether the intermediate representation is consistent when parsed from different inputs.

    IR is a dictionary of form:
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    """

    def test_from_sqlalchemy_table(self) -> None:
        """
        Tests that `parse.sqlalchemy_table` produces `intermediate_repr_no_default_sql_doc` properly
        """

        # Sanity check
        run_ast_test(
            self,
            config_tbl_with_comments_ast,
            gold=ast.parse(config_tbl_with_comments_str).body[0],
        )

        for variant in (
            config_tbl_with_comments_str,
            config_tbl_with_comments_str.replace(
                "config_tbl =", "config_tbl: Table =", 1
            ),
            config_tbl_with_comments_str.replace("config_tbl =", "", 1).lstrip(),
        ):
            ir = cdd.sqlalchemy.parse.sqlalchemy_table(ast.parse(variant).body[0])
            self.assertEqual(ir["name"], "config_tbl")
            ir["name"] = None
            self.assertDictEqual(ir, intermediate_repr_no_default_sql_with_sql_types)

    def test_from_sqlalchemy(self) -> None:
        """
        Tests that `parse.sqlalchemy` produces `intermediate_repr_no_default_sql_doc` properly
        """

        # Sanity check
        run_ast_test(
            self,
            config_decl_base_ast,
            gold=ast.parse(config_decl_base_str).body[0],
        )

        ir = cdd.sqlalchemy.parse.sqlalchemy(deepcopy(config_decl_base_ast))
        self.assertEqual(ir["name"], "config_tbl")
        ir["name"] = None
        self.assertDictEqual(ir, intermediate_repr_no_default_sql_with_sql_types)

    def test_from_sqlalchemy_with_foreign_rel(self) -> None:
        """Test from SQLalchemy with a foreign key relationship"""
        # Sanity check
        run_ast_test(
            self,
            foreign_sqlalchemy_tbls_mod,
            gold=ast.parse(foreign_sqlalchemy_tbls_str),
        )
        ir = cdd.sqlalchemy.parse.sqlalchemy_table(deepcopy(node_pk_tbl_ass))
        self.assertDictEqual(ir, intermediate_repr_node_pk)


unittest_main()



"""
Parser utils for routes
"""

from ast import Call, FunctionDef

from cdd.routes.parse.bottle import methods
from cdd.shared.ast_utils import get_value
from cdd.shared.pure_utils import rpartial


def get_route_meta(mod):
    """
    Get the (func_name, app_name, route_path, http_method)s

    :param mod: Parsed AST containing routes
    :type mod: ```Module```

    :return: Iterator of tuples of (func_name, app_name, route_path, http_method)
    :rtype: ```Iterator[Tuple[str, str, str, str]]```
    """
    return map(
        lambda func: (
            func.name,
            *next(
                map(
                    lambda call: (
                        call.func.value.id,
                        get_value(call.args[0]),
                        call.func.attr,
                    ),
                    filter(
                        lambda call: call.args and call.func.attr in methods,
                        filter(rpartial(isinstance, Call), func.decorator_list),
                    ),
                )
            ),
        ),
        filter(rpartial(isinstance, FunctionDef), mod.body),
    )


__all__ = ["get_route_meta"]



"""
Utility functions for `cdd.emit.argparse_function`
"""

import ast
from ast import Name, Return
from typing import Any

from cdd.shared.ast_utils import NoneStr, get_value, set_value
from cdd.shared.defaults_utils import extract_default, set_default_doc
from cdd.shared.pure_utils import identity, simple_types
from cdd.shared.source_transformer import to_code


def _parse_return(e, intermediate_repr, function_def, emit_default_doc):
    """
    Parse return into a param dict

    :param e: Return AST node
    :type e: Return

    :param intermediate_repr: a dictionary of form
        {  "name": Optional[str],
           "type": Optional[str],
           "doc": Optional[str],
           "params": OrderedDict[str, {'typ': str, 'doc': Optional[str], 'default': Any}]
           "returns": Optional[OrderedDict[Literal['return_type'],
                                           {'typ': str, 'doc': Optional[str], 'default': Any}),)]] }
    :type intermediate_repr: ```dict```

    :param function_def: AST node for function definition
    :type function_def: ```FunctionDef```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Name, dict with keys: 'typ', 'doc', 'default'
    :rtype: ```Tuple[str, dict]```
    """
    assert isinstance(e, Return), "Expected `Return` got `{type_name}`".format(
        type_name=type(e).__name__
    )

    typ = intermediate_repr["returns"]["return_type"]["typ"]
    if "[" in intermediate_repr["returns"]["return_type"]["typ"]:
        typ = to_code(get_value(ast.parse(typ).body[0].value.slice).elts[1]).rstrip(
            "\n"
        )

    return set_default_doc(
        (
            "return_type",
            {
                "doc": extract_default(
                    next(
                        line.partition(",")[2].lstrip()
                        for line in get_value(function_def.body[0].value).split("\n")
                        if line.lstrip().startswith(":return")
                    ),
                    emit_default_doc=emit_default_doc,
                )[0],
                "default": to_code(e.value.elts[1]).rstrip("\n"),
                "typ": typ
                # 'Tuple[ArgumentParser, {typ}]'.format(typ=intermediate_repr['returns']['typ'])
            },
        ),
        emit_default_doc=emit_default_doc,
    )


def parse_out_param(expr, require_default=False, emit_default_doc=True):
    """
    Turns the class_def repr of '--dataset_name', type=str, help='name of dataset.', required=True, default='mnist'
      into
           Tuple[Literal['dataset_name'], {"typ": Literal["str"], "doc": Literal["name of dataset."],
                                           "default": Literal["mnist"]}]

    :param expr: Expr
    :type expr: ```Expr```

    :param require_default: Whether a default is required, if not found in doc, infer the proper default from type
    :type require_default: ```bool```

    :param emit_default_doc: Whether help/docstring should include 'With default' text
    :type emit_default_doc: ```bool```

    :return: Name, dict with keys: 'typ', 'doc', 'default'
    :rtype: ```Tuple[str, dict]```
    """
    # print("require_default:", require_default, ";")
    required = get_value(
        get_value(
            next(
                (
                    keyword
                    for keyword in expr.value.keywords
                    if keyword.arg == "required"
                ),
                set_value(False),
            )
        )
    )

    typ = next(
        (
            _handle_value(get_value(key_word))
            for key_word in expr.value.keywords
            if key_word.arg == "type"
        ),
        "str",
    )
    name = get_value(expr.value.args[0])[len("--") :]
    default = next(
        (
            get_value(key_word.value)
            for key_word in expr.value.keywords
            if key_word.arg == "default"
        ),
        None,
    )
    doc = (
        lambda help_: help_
        if help_ is None
        else (
            help_
            if default is None
            or emit_default_doc is False
            or (hasattr(default, "__len__") and len(default) == 0)
            or "defaults to" in help_
            or "Defaults to" in help_
            else "{help} Defaults to {default}".format(
                help=help_ if help_.endswith(".") else "{}.".format(help_),
                default=default,
            )
        )
    )(
        next(
            (
                get_value(key_word.value)
                for key_word in expr.value.keywords
                if key_word.arg == "help" and key_word.value
            ),
            None,
        )
    )
    if default is None:
        doc, default = extract_default(doc, emit_default_doc=emit_default_doc)
    if default is None:
        if required:
            # if name.endswith("kwargs"):
            #    default = NoneStr
            # else:
            default = simple_types[typ] if typ in simple_types else NoneStr

        elif require_default:  # or typ.startswith("Optional"):
            default = NoneStr

    action = next(
        (
            get_value(key_word.value)
            for key_word in expr.value.keywords
            if key_word.arg == "action"
        ),
        None,
    )

    typ = next(
        (
            _handle_keyword(keyword, typ)
            for keyword in expr.value.keywords
            if keyword.arg == "choices"
        ),
        typ,
    )
    if action == "append":
        typ = "List[{typ}]".format(typ=typ)

    if not required and "Optional" not in typ:
        typ = "Optional[{typ}]".format(typ=typ)

    return name, dict(
        doc=doc, typ=typ, **({} if default is None else {"default": default})
    )


def _handle_keyword(keyword, typ):
    """
    Decide which type to wrap the keyword tuples in

    :param keyword: AST keyword
    :type keyword: ```ast.keyword```

    :param typ: string representation of type
    :type typ: ```str```

    :return: string representation of type
    :rtype: ```str```
    """
    quote_f = identity

    type_ = "Union"
    if typ == Any or typ in simple_types:
        if typ in ("str", Any):

            def quote_f(s):
                """
                Wrap the input in quotes

                :param s: Any value
                :type s: ```Any```

                :return: the input value
                :rtype: ```Any```
                """
                return "'{}'".format(s)

        type_ = "Literal"

    return "{type}[{types}]".format(
        type=type_,
        types=", ".join(quote_f(get_value(elt)) for elt in keyword.value.elts),
    )


def _handle_value(node):
    """
    Handle keyword.value types, returning the correct one as a `str` or `Any`

    :param node: AST node from keyword.value
    :type node: ```Name```

    :return: `str` or `Any`, representing the type for argparse
    :rtype: ```Union[str, Any]```
    """
    # if isinstance(node, Attribute): return Any
    if isinstance(node, Name):
        return "Optional[dict]" if node.id == "loads" else node.id
    raise NotImplementedError(type(node).__name__)


__all__ = ["_parse_return", "parse_out_param", "_handle_keyword", "_handle_value"]



"""
Shared by the mocks. Currently unused, but has some imports mocked for later use…
"""

from ast import parse as ast_parse

from cdd.shared.pure_utils import PY_GTE_3_8

imports_header = """
from {package} import Literal
from typing import Optional, Tuple, Union

try:
    import tensorflow as tf
    import numpy as np
except ImportError:
    tf = type('TensorFlow', tuple(), {{ 'data': type('Dataset', tuple(), {{ "Dataset": None }}) }} )
    np = type('numpy', tuple(), {{ 'ndarray': None, 'empty': lambda _: _ }})
""".format(
    package="typing" if PY_GTE_3_8 else "typing_extensions"
)

imports_header_ast = ast_parse(imports_header).body

__all__ = ["imports_header", "imports_header_ast"]



""" Tests for CLI exmod subparser (__main__.py) """

from unittest import TestCase
from unittest.mock import MagicMock, patch

from cdd.tests.utils_for_tests import run_cli_test, unittest_main


class TestCliExMod(TestCase):
    """Test class for __main__.py"""

    def test_exmod_fails(self) -> None:
        """Tests CLI interface exmod failure cases"""
        run_cli_test(
            self,
            ["exmod", "--wrong"],
            exit_code=2,
            output="the following arguments are required: --module/-m, --emit, --output-directory/-o\n",
        )

    def test_exmod_is_called(self) -> None:
        """Tests CLI interface exmod function gets called"""
        with patch("cdd.__main__.exmod", new_callable=MagicMock()):
            self.assertTrue(
                run_cli_test(
                    self,
                    [
                        "exmod",
                        "--module",
                        "foo",
                        "--emit",
                        "argparse",
                        "--output-directory",
                        "foo",
                    ],
                    exit_code=None,
                    output=None,
                ),
            )


unittest_main()



"""
Utility functions for `cdd.parse.class`
"""

from inspect import getsource


def get_source(obj):
    """
    Call inspect.getsource and raise an error unless class definition could not be found

    :param obj: object to inspect
    :type obj: ```Any```

    :return: The source
    :rtype: ```str```
    """
    try:
        return getsource(obj)
    except OSError as e:
        if e.args and e.args[0] in frozenset(
            (
                "could not find class definition",
                "source code not available",
                "could not get source code",
            )
        ):
            return None
        raise


__all__ = ["get_source"]



""" Tests for CLI gen_routes subparser (__main__.py) """

from unittest import TestCase
from unittest.mock import patch

from cdd.tests.utils_for_tests import mock_function, run_cli_test, unittest_main


class TestCliGenRoutes(TestCase):
    """Test class for __main__.py"""

    def test_gen_routes_fails(self) -> None:
        """Tests CLI interface failure cases"""
        run_cli_test(
            self,
            ["gen_routes", "--wrong"],
            exit_code=2,
            output="the following arguments are required: --crud, --model-path, --model-name, --routes-path\n",
        )

    def test_gen_routes(self) -> None:
        """Tests CLI interface gets all the way to the gen_routes call without error"""

        with patch(
            "cdd.__main__.gen_routes", lambda *args, **kwargs: (True,) * 2
        ), patch("cdd.__main__.upsert_routes", mock_function):
            self.assertTrue(
                run_cli_test(
                    self,
                    [
                        "gen_routes",
                        "--crud",
                        "CRD",
                        "--model-path",
                        "cdd.tests.mocks.sqlalchemy",
                        "--routes-path",
                        "/api/config",
                        "--model-name",
                        "Config",
                    ],
                    exit_code=None,
                    output=None,
                ),
            )


unittest_main()



"""
Utility functions for `cdd.parse.openapi`
"""


def extract_entities(openapi_str):
    """
    Extract entities from an OpenAPI string, where entities are defines as anything within "```"

    :param openapi_str: The OpenAPI str
    :type openapi_str: ```str```

    :return: Entities
    :rtype: ```List[str]```
    """
    entities, ticks, space, stack = [], 0, 0, []

    def add_then_clear_stack():
        """
        Join entity, if non empty add to entities. Clear stack.
        """
        entity = "".join(stack)
        if entity:
            entities.append(entity)
        stack.clear()

    for idx, ch in enumerate(openapi_str):
        if ch.isspace():
            space += 1
            add_then_clear_stack()
            ticks = 0
        elif ticks > 2:
            ticks, space = 0, 0
            stack and add_then_clear_stack()
            stack.append(ch)
        elif ch == "`":
            ticks += 1
        elif stack:
            stack.append(ch)
    add_then_clear_stack()
    return entities


__all__ = ["extract_entities"]



"""
Utils for working with AST (builtin) and cdd's CST
"""

from copy import deepcopy
from enum import Enum
from itertools import takewhile
from operator import attrgetter, ne
from sys import stderr

from cdd.shared.ast_utils import cmp_ast, get_doc_str
from cdd.shared.cst_utils import (
    FunctionDefinitionStart,
    TripleQuoted,
    UnchangingLine,
    ast2cst,
)
from cdd.shared.pure_utils import count_iter_items, omit_whitespace, tab
from cdd.shared.source_transformer import to_code


def debug_doctrans(changed, affector, name, typ):
    """
    Print debug statement if changed is not nop

    :param changed: Delta value indicating what changed (if anything)
    :type changed: ```Delta```

    :param affector: What is being changed
    :type affector: ```str```

    :param name: Name of what is being changed
    :type name: ```str```

    :param typ: AST type name of what is being changed
    :type typ: ```str```
    """
    if changed is not Delta.nop:
        print(
            "{changed!s}".format(changed=changed).ljust(20),
            "{affector}\t{typ}\t`{name}`".format(affector=affector, typ=typ, name=name),
            sep="",
        )


def find_cst_at_ast(cst_list, node):
    """
    Find (first) CST node matching AST node

    (uses `_location` from `annotate_ancestry`)

    :param cst_list: List of `namedtuple`s with at least ("line_no_start", "line_no_end", "value") attributes
    :type cst_list: ```List[NamedTuple]```

    :param node: AST node
    :type node: ```AST```

    :return: Matching idx and element from cst_list if found else (None, None)
    :rtype: ```Tuple[Optional[int], Optional[NamedTuple]]````
    """
    cst_node_found, cst_node_no = None, None
    node_type = type(node).__name__
    cst_type = ast2cst.get(node_type, type(None)).__name__
    if cst_type == "NoneType":
        print("`{node_type}` not implemented".format(node_type=node_type), file=stderr)
        return None, None
    for cst_node_no, cst_node in enumerate(cst_list):
        if (
            cst_node.line_no_start <= node.lineno <= cst_node.line_no_end
            # Extra precautions to ensure the wrong new_node is never replaced:
            and type(cst_node).__name__ == cst_type  # `isinstance` doesn't work
            and getattr(cst_node, "name", None) == getattr(node, "name", None)
        ):
            cst_node_found = cst_node
            break
    return cst_node_no, cst_node_found


class Delta(Enum):
    """
    Maybe Enum for what every `maybe_` function in `ast_cst_utils` can return
    """

    added = 0
    removed = 1
    replaced = 2
    nop = 255


def maybe_replace_doc_str_in_function_or_class(node, cst_idx, cst_list):
    """
    Maybe replace the doc_str of a function or class

    :param node: AST node
    :type node: ```Union[ClassDef, AsyncFunctionDef, FunctionDef]```

    :param cst_idx: Index of start of function/class in cst_list
    :type cst_idx: ```int```

    :param cst_list: List of `namedtuple`s with at least ("line_no_start", "line_no_end", "value") attributes
    :type cst_list: ```List[NamedTuple]```

    :return: Delta value indicating what changed (if anything)
    :rtype: ```Delta```
    """
    new_doc_str = get_doc_str(node) or ""
    cur_node_after_func = (
        cst_list[cst_idx + 1]
        if cst_idx + 1 < len(cst_list)
        else UnchangingLine(0, 0, "")
    )
    existing_doc_str = (
        isinstance(cur_node_after_func, TripleQuoted) and cur_node_after_func.is_docstr
    )
    changed = Delta.nop

    def formatted_doc_str(doc_str, is_double_q=True):
        """
        Correctly indent, pre and post space the doc_str

        :param doc_str: Input doc string
        :type doc_str: ```str```

        :param is_double_q: Whether the doc_str should be double quoted
        :type is_double_q: ```bool```

        :return: Correctly formatted `doc_str`
        :rtype: ```str```
        """
        str_after_func_no_nl = cur_node_after_func.value.lstrip("\n")
        indent_after_func_no_nl = count_iter_items(
            takewhile(str.isspace, str_after_func_no_nl)
        )
        space = str_after_func_no_nl[:indent_after_func_no_nl]
        return TripleQuoted(
            is_double_q=is_double_q,
            is_docstr=True,
            value='\n{space}"""{replacement_doc_str}\n{space}"""'.format(
                space=space,
                replacement_doc_str="\n".join(
                    map(
                        lambda line: "{space}{line}".format(
                            space=str_after_func_no_nl[
                                : indent_after_func_no_nl - len(tab)
                            ],
                            line=line,
                        ),
                        doc_str.split("\n"),
                    )
                ).rstrip(),
            ),
            line_no_start=cur_node_after_func.line_no_start,
            line_no_end=cur_node_after_func.line_no_end,
        )

    if new_doc_str and not existing_doc_str:
        cst_list.insert(
            cst_idx + 1,
            formatted_doc_str(new_doc_str),
        )
        changed = Delta.added
    elif not new_doc_str and existing_doc_str:
        del cst_list[cst_idx + 1]
        changed = Delta.removed
    # elif not new_doc_str and not existing_doc_str: changed = Delta.nop
    elif new_doc_str and existing_doc_str:
        cur_doc_str_only = cur_node_after_func.value.strip()[3:-3]
        if ne(*map(omit_whitespace, (cur_doc_str_only, new_doc_str))):
            pre, _, post = cur_node_after_func.value.partition(cur_doc_str_only)
            cst_list[cst_idx + 1] = formatted_doc_str(
                new_doc_str, is_double_q=cst_list[cst_idx + 1].is_double_q
            )
            changed = Delta.replaced
    if changed is not Delta.nop:
        debug_doctrans(changed, "docstr", node.name, type(node).__name__)
        # Subsequent `line_no` `start,end` lines are invalidated. It's necessary to link the CST and AST together.

    return changed


def maybe_replace_function_return_type(new_node, cur_ast_node, cst_idx, cst_list):
    """
    Maybe replace the function's return type

    :param new_node: AST function node
    :type new_node: ```Union[AsyncFunctionDef, FunctionDef]```

    :param cur_ast_node: AST function node of CST (with fake body)
    :type cur_ast_node: ```AST```

    :param cst_idx: Index of start of function/class in cst_list
    :type cst_idx: ```int```

    :param cst_list: List of `namedtuple`s with at least ("line_no_start", "line_no_end", "value") attributes
    :type cst_list: ```List[NamedTuple]```

    :return: Delta value indicating what changed (if anything)
    :rtype: ```Delta```
    """
    new_node = deepcopy(new_node)
    new_node.body = cur_ast_node.body
    value = None

    def remove_return_typ(statement):
        """
        Remove the return typ

        :param statement: The statement verbatim
        :type statement: ```str```

        :return: The new function prototype
        :rtype: ```str```
        """
        return "{type_less}:".format(
            type_less=statement[: statement.rfind("->")].rstrip()
        )

    def add_return_typ(statement):
        """
        Add the return typ

        :param statement: The statement verbatim
        :type statement: ```str```

        :return: The new function prototype
        :rtype: ```str```
        """
        pre, col, post = statement.rpartition(":")
        return "{pre} -> {return_typ}{col}{post}".format(
            pre=pre,
            return_typ=to_code(new_node.returns).rstrip("\n"),
            col=col,
            post=post,
        )

    if cmp_ast(cur_ast_node.returns, new_node.returns):
        changed = Delta.nop
    elif cur_ast_node.returns and new_node.returns:
        changed = Delta.replaced
        value = add_return_typ(remove_return_typ(cst_list[cst_idx].value))
    elif cur_ast_node.returns and not new_node.returns:
        changed = Delta.removed
        value = remove_return_typ(cst_list[cst_idx].value)
    else:  # not cur_ast_node.returns and new_node.returns:
        changed = Delta.added
        value = add_return_typ(cst_list[cst_idx].value)
    if value is not None:
        cst_list[cst_idx] = FunctionDefinitionStart(
            line_no_start=cst_list[cst_idx].line_no_start,
            line_no_end=cst_list[cst_idx].line_no_end,
            name=cst_list[cst_idx].name,
            value=value,
        )

    if changed is not Delta.nop:
        debug_doctrans(changed, "return_type", new_node.name, type(new_node).__name__)

    return changed


def maybe_replace_function_args(new_node, cur_ast_node, cst_idx, cst_list):
    """
    Maybe replace the doc_str of a function or class

    :param new_node: AST function node
    :type new_node: ```Union[AsyncFunctionDef, FunctionDef]```

    :param cur_ast_node: AST function node of CST (with fake body)
    :type cur_ast_node: ```AST```

    :param cst_idx: Index of start of function/class in cst_list
    :type cst_idx: ```int```

    :param cst_list: List of `namedtuple`s with at least ("line_no_start", "line_no_end", "value") attributes
    :type cst_list: ```List[NamedTuple]```

    :return: Delta value indicating what changed (if anything)
    :rtype: ```Delta```
    """
    new_node = deepcopy(new_node)
    new_node.body = cur_ast_node.body
    changed = Delta.nop
    if not cmp_ast(cur_ast_node.args, new_node.args):
        new_args, cur_args = map(attrgetter("args.args"), (new_node, cur_ast_node))

        for i in range(len(cur_args)):
            if cur_args[i].annotation != new_args[i].annotation:
                # Approximation, obviously you could have intermixed annotation and to-be (un)annotated
                if cur_args[i].annotation is None:
                    changed = Delta.added
                elif new_args[i].annotation is None:
                    changed = Delta.removed
                else:
                    changed = Delta.replaced
                break

        pre, returning, post = cst_list[cst_idx].value.rpartition("->")
        left_parens, right_parens = 0, 0
        start_idx, end_idx = pre.rfind("("), pre.rfind(")")
        for start_idx in range(end_idx, 0, -1):
            if pre[start_idx] == ")":
                right_parens += 1
            elif pre[start_idx] == "(":
                left_parens += 1

            if right_parens == left_parens and right_parens > 0:
                break

        cst_list[cst_idx] = FunctionDefinitionStart(
            line_no_start=cst_list[cst_idx].line_no_start,
            line_no_end=cst_list[cst_idx].line_no_end,
            name=cst_list[cst_idx].name,
            # TODO: Handle comments in the middle of args, and match whitespace, and maybe even limit line length
            value="{start}{args}{end}{returning}{post}".format(
                start=pre[: start_idx + 1],
                args=", ".join(
                    "{arg_name}{annotation}".format(
                        annotation=""
                        if arg.annotation is None
                        else ": {annotation_unparsed}".format(
                            annotation_unparsed=to_code(arg.annotation).rstrip("\n")
                        ),
                        arg_name=arg.arg,
                    )
                    for arg in new_args
                ),
                end=pre[end_idx:],
                returning=returning,
                post=post,
            ),
        )

    if changed is not Delta.nop:
        debug_doctrans(changed, "args", new_node.name, type(new_node).__name__)

    return changed


__all__ = [
    "Delta",
    "debug_doctrans",
    "find_cst_at_ast",
    "maybe_replace_doc_str_in_function_or_class",
    "maybe_replace_function_args",
    "maybe_replace_function_return_type",
]
